/******************************************************************************
 * commandman.cpp - Command Manager (the editor)
 *
 * yamce - Yet Another Minecraft Editor
 * Copyright(c) 2019 - Dick Balaska, and BuckoSoft.
 *
 * yamce is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * yamce is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/
#include <QDebug>
#include <QAction>

#include "commandman.h"
#include "command.h"
#include "nbtfile.h"
#include "resourcedock.h"
#include "treemodel.h"
#include "mainwindow.h"
#include "mapwidget.h"
#include "mapfile.h"

CommandMan::CommandMan(MainWindow* mainWindow)
	: QObject (mainWindow),
	  m_mainWindow(mainWindow)
{
}

CommandMan::~CommandMan()
{
	while (!m_undoList.isEmpty())
		delete m_undoList.takeLast();
	while (!m_redoList.isEmpty()) {
		Command* c = m_redoList.takeLast();
		c->disown();
		delete c;
	}
}

ResourceDock* CommandMan::getResourceDock()
{
	return(m_mainWindow->getResourceDock());
}

NBTFile* CommandMan::getFile(TreeItem *treeItem)
{
	return(getResourceDock()->getFile(treeItem));
}

void CommandMan::setUndoRedo(QAction* undo, QAction* redo, QAction* save, QAction* sync)
{
	m_undoAction = undo;
	m_redoAction = redo;
	m_saveAction = save;
	m_syncAction = sync;
	enableUndoRedo();
}

void CommandMan::enableUndoRedo()
{
	bool b = false;
	QString type;
	if (!m_undoList.isEmpty()) {
		b = true;
		type = m_undoList.last()->getName();
	}
	m_undoAction->setText(QString("%1 %2").arg(tr("Undo"), type));
	m_undoAction->setEnabled(b);
	m_saveAction->setEnabled(b);
	QString s;
	if (b)
		s = "*";
	if (m_mainWindow->getWorld())
		s = QString("%1%2 - yamce").arg(m_mainWindow->getWorld()->getName(), s);
	else
		s = "yamce";
	m_mainWindow->setWindowTitle(s);

	b = false;
	type.clear();
	if (!m_redoList.isEmpty()) {
		b = true;
		type = m_redoList.last()->getName();
	}
	m_redoAction->setText(QString("%1 %2").arg(tr("Redo"), type));
	m_redoAction->setEnabled(b);

	m_syncAction->setEnabled(m_undoList.isEmpty());
}

void CommandMan::markDirtyOrClean(NBTFile* file)
{
	bool dirty = false;
	for (Command* c : m_undoList) {
		if (c->m_file == file) {
			dirty = true;
			break;
		}
	}
	Group* g = m_mainWindow->getWorld()->findGroup(file);
	TreeModel* treeModel = g->getTreeModel();
	QString cleanName = file->getFileName();
	QString dirtyName = QString("%1*").arg(cleanName);
	QModelIndex qmi;
	bool fileCurrentlyDirty = false;
	for (int i=0; i<treeModel->rowCount(); i++) {
		QModelIndex qmix = treeModel->index(i, 0);
		QVariant data = treeModel->data(qmix);
		QString s = data.toString();
		fileCurrentlyDirty = false;
		if (s.endsWith('*')) {
			s = s.left(s.length()-1);
			fileCurrentlyDirty = true;
		}
		if (s == cleanName) {
			qmi = qmix;
			//fileCurrentlyDirty = false;
			break;
		}
	}
	if (!qmi.isValid()) {
		qWarning() << "CommandMan:markDirtyOrClean can't find file in tree" << cleanName;
		return;
	}
	if (dirty) {
		file->setDirty(true);
		if (!fileCurrentlyDirty) {
			QMap<int, QVariant> data;
			data[Qt::DisplayRole] = QVariant(dirtyName);
			treeModel->setItemData(qmi, data);
		}
	} else {
		file->setDirty(false);
		if (fileCurrentlyDirty) {
			QMap<int, QVariant> data;
			data[Qt::DisplayRole] = QVariant(cleanName);
			treeModel->setItemData(qmi, data);
		}
	}
}

void CommandMan::worldSaved()
{
	QList<NBTFile*> files;
	while (!m_undoList.isEmpty()) {
		Command* c = m_undoList.takeLast();
		if (!files.contains(c->m_file))
			files.append(c->m_file);
		delete c;
	}
	while (!m_redoList.isEmpty()) {
		Command* c = m_redoList.takeLast();
		c->disown();
		delete c;
	}
	for (NBTFile* file : files) {
		markDirtyOrClean(file);
	}
	enableUndoRedo();
}

void CommandMan::postCommand(Command* c)
{
	enableUndoRedo();
	markDirtyOrClean(c->m_file);
	MapFile* mf = dynamic_cast<MapFile*>(c->m_file);
	if (mf) {
		mf->parseMap();
		if (m_mainWindow->getResourceDock()->getCurrentGroupIndex() == wgMaps)
			m_mainWindow->getMapMapWidget()->fileLoaded(mf);
	}
}

void CommandMan::doCommand(Command* command)
{
	command->m_commandMan = this;
	m_undoList.append(command);
	command->doCommand();
	postCommand(command);
}

void CommandMan::undo()
{
	Command* c = m_undoList.takeLast();
	if (DEBUGCOMMAND)
		qDebug() << "CommandMan:undo" << c->getName();
	m_redoList.append(c);
	c->undoCommand();
	postCommand(c);
}

void CommandMan::redo()
{
	Command* c = m_redoList.takeLast();
	if (DEBUGCOMMAND)
		qDebug() << "CommandMan:redo" << c->getName();
	m_undoList.append(c);
	c->redoCommand();
	postCommand(c);
}

