/******************************************************************************
 * consoleman.cpp - Manage the app messages
 *
 * yamce - Yet Another Minecraft Editor
 * Copyright(c) 2019 - Dick Balaska, and BuckoSoft.
 *
 * yamce is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * yamce is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/
#include <QTime>
#include <QTextEdit>
#include <QMutex>

#include "consoledialog.h"
#include "consoleman.h"
#include "yamcedoc.h"
#include "mainwindow.h"

static const char*	msgTypeNames[] = {"DEBG", "WARN", "CRIT", "FATL", "INFO"};

ConsoleMan* globalConsoleMan = nullptr;

ConsoleMan::ConsoleMan(MainWindow* mainWindow)
	: m_mainWindow(mainWindow),
	  m_consoleDialog(nullptr),
	  m_consoleDialogSettings(nullptr)
{
	globalConsoleMan = this;
	m_showDebug = true;
}

void ConsoleMan::setWorld(World* world)
{
	m_world = world;
}

void ConsoleMan::onConsoleDialog(bool)

{
	if (!m_consoleDialog) {
		m_consoleDialog = new ConsoleDialog(m_mainWindow, m_mainWindow);
		connect(m_consoleDialog, SIGNAL(rejected()), this, SLOT(onConsoleDialogClosed()));
		connect(m_consoleDialog, SIGNAL(accepted()), this, SLOT(onConsoleDialogClosed()));

		for (ConsoleMsg* cm : m_messageQueue) {
			m_consoleDialog->appendMsg(cm);
		}
	}
	if (m_consoleDialogSettings && !m_consoleDialogSettings->m_pos.isNull()) {
		m_consoleDialog->move(m_consoleDialogSettings->m_pos.topLeft());
		m_consoleDialog->resize(m_consoleDialogSettings->m_pos.size());
	}
	m_consoleDialog->raise();
	m_consoleDialog->show();
}

void ConsoleMan::setupConsoleDialog(WindowSettings* settings)
{
	m_consoleDialogSettings = settings;
	if (settings->m_shown) {
		onConsoleDialog(true);
	}
}

void ConsoleMan::updateConsoleDialog()
{
	if (m_consoleDialogSettings) {
		if (m_consoleDialog) {
			m_consoleDialogSettings->m_pos.setTopLeft(m_consoleDialog->pos());
			m_consoleDialogSettings->m_pos.setSize(m_consoleDialog->size());
		}
		m_consoleDialogSettings->m_shown = (m_consoleDialog != nullptr);
	}
}

void  ConsoleMan::resetConsoleDialog()
{
	m_consoleDialogSettings = nullptr;
	if (m_consoleDialog)
		m_consoleDialog->close();
}


void ConsoleMan::onConsoleDialogClosed()
{
	updateConsoleDialog();
	m_consoleDialog = nullptr;
	if (m_consoleDialogSettings)
		m_consoleDialogSettings->m_shown = false;
}

QMutex consoleMutex;

void ConsoleMan::consoleMessage(QtMsgType type, const QMessageLogContext &context, const QString &msg)
{
	if (type == QtDebugMsg && !m_showDebug)
		return;
	QString file = "";
	if (context.line != 0)
		file = QString(" (%1:%2)").arg(context.file).arg(context.line);
	QString s = QString("[%1] %2 %3%4")
			.arg(QTime::currentTime().toString("hh:mm:ss.zzz"), msgTypeNames[type], msg, file);
	ConsoleMsg* cm = new ConsoleMsg();
	cm->m_type = type;
	cm->m_string = s;
	m_messageQueue.append(cm);
	if (m_consoleDialog) {
		consoleMutex.lock();
		m_consoleDialog->appendMsg(cm);
		consoleMutex.unlock();
	}
	QByteArray ba = s.toLocal8Bit();
	fprintf(stderr, "%s\n", ba.constData());
}
