/******************************************************************************
 * mainwindow.cpp - The standard Qt MainWindow
 *
 * yamce - Yet Another Minecraft Editor
 * Copyright(c) 2018 - Dick Balaska, and BuckoSoft.
 *
 * yamce is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * yamce is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/
#include <QDebug>
#include <QSettings>
#include <QToolBar>
#include <QFileDialog>
#include <QMessageBox>

#include "resourcedock.h"
#include "regionwidget.h"
#include "centralwidget.h"
#include "consoleman.h"
#include "helpman.h"
#include "mapwidget.h"
#include "nbtobject.h"
#include "world.h"
#include "maintoolbar.h"
#include "statusbar.h"
#include "fileman.h"
#include "commandman.h"
#include "syncman.h"
#include "regionfile.h"
#include "mapfile.h"
#include "yamcedoc.h"
#include "railman.h"
#include "mapbannerpainter.h"
#include "preferencesdialog.h"
#include "mainwindow.h"

static QString   s_companyName = "BuckoSoft";
static QString   s_productName = "yamce";

static bool DEBUG = false;

QPixmap* logoPixmap;

MainWindow::MainWindow(QWidget *parent)
	: QMainWindow(parent),
	  m_world(nullptr),
	  m_prefsDialog(nullptr)
{
	NBTObject::init();
	MapBannerPainter::init();

	logoPixmap = new QPixmap(":/resources/128x128/yamce.png");
	setWindowIcon(QIcon(*logoPixmap));
	m_commandMan = new CommandMan(this);
	m_consoleMan = new ConsoleMan(this);
	m_helpMan = new HelpMan(this);
	m_syncMan = new SyncMan(this);

	m_resourceDock = new ResourceDock(this);
	this->addDockWidget(Qt::LeftDockWidgetArea, m_resourceDock);
	m_centralWidget = new CentralWidget(this);
	m_regionMapScrollArea = m_centralWidget->m_regionWidget->m_regionMapScrollArea;
	m_regionMapWidget = m_regionMapScrollArea->m_widget;
	m_mapWidget = m_centralWidget->m_mapWidget;
	m_mapMapWidget = m_mapWidget->getMapMapWidget();
	this->setCentralWidget(m_centralWidget);

	m_mainToolbar = new MainToolbar(24, this);
	addToolBar(m_mainToolbar);
	m_statusBar = new StatusBar(this);
	this->setStatusBar(m_statusBar);

	m_fileMan = new FileMan(this);
	m_railMan = new RailMan(this);
	m_railMan->setFileMan(m_fileMan);


	connect(m_fileMan, SIGNAL(readFileDone(NBTFile*)), this, SLOT(fileLoaded(NBTFile*)));
	connect(m_fileMan, SIGNAL(readFileDone(NBTFile*)), m_resourceDock, SLOT(fileLoaded(NBTFile*)));
	connect(m_fileMan, SIGNAL(readFileDone(NBTFile*)), m_centralWidget, SLOT(fileLoaded(NBTFile*)));
	connect(m_fileMan, SIGNAL(readFileDone(NBTFile*)), m_railMan, SLOT(fileLoaded(NBTFile*)));
	connect(m_fileMan, SIGNAL(sendErrorMsg(QString)), this, SLOT(errorMsgReceived(QString)));
	connect(m_regionMapWidget, SIGNAL(selectRegion(int, int)), this, SLOT(regionChanged(int, int)));
	connect(this, SIGNAL(changeZoomValue(double)), m_regionMapScrollArea, SLOT(zoomValueChanged(double)));
	connect(this, SIGNAL(changeZoomValue(double)), m_mapMapWidget, SLOT(onZoomChanged(double)));
	connect(this, SIGNAL(changeCurrentResourceGroup(int)), m_resourceDock, SLOT(explorerTypeChanged(int)));
	connect(this, SIGNAL(changeCurrentResourceGroup(int)), m_regionMapScrollArea, SLOT(explorerTypeChanged(int)));
	connect(this, SIGNAL(changeCurrentResourceGroup(int)), m_centralWidget, SLOT(explorerTypeChanged(int)));
	connect(this, SIGNAL(doExportPNG()), m_mapMapWidget, SLOT(exportPNG()));
	connect(m_mapWidget, SIGNAL(processRailsChecked()), m_railMan, SLOT(buildRailPoints()));
	connect(m_railMan, SIGNAL(railManDone()), m_mapWidget, SLOT(onRailManDone()));
	connect(m_railMan, SIGNAL(railManDone()), m_fileMan, SLOT(onRailManDone()));
	connect(m_mapMapWidget, SIGNAL(mapSelected(MapFile*)), m_resourceDock, SLOT(setMapFile(MapFile*)));
	connect(m_mainToolbar, SIGNAL(zoomValueChanged(double)), this, SLOT(onZoomValueChanged(double)));

	loadPreferences();	// loadPreferences is late because it triggers signals (sets combobox values)

	bool loaded = false;
	if (!m_currentDirectory.isEmpty())
		loaded = loadWorld();
	if (!loaded) {
		openWorld();
	}
}

MainWindow::~MainWindow()
{
	savePreferences();
	delete m_helpMan;
	delete m_syncMan;
	delete m_fileMan;
	delete m_railMan;
	delete m_commandMan;
	delete m_world;
}

void MainWindow::openWorld()
{
	if (DEBUG)
		qDebug() << "openWorld";
	QString dir = QString("%1/.minecraft/saves").arg(QDir::homePath());
	m_currentDirectory = QFileDialog::getExistingDirectory(this, tr("Select Directory"), dir, QFileDialog::ShowDirsOnly);
	if (!m_currentDirectory.isEmpty()) {
		loadWorld();
	}
}

bool MainWindow::loadWorld()
{
	World* w = new World(this);
	bool res = w->load(m_currentDirectory, this);
	if (!res)
		return(false);
	w->m_yamceDoc.load(this, w);
	if (m_world) {
		captureWidgets();
		resetWidgets();
		m_syncMan->onSynchronize(false);
		m_world->m_yamceDoc.save(this, m_world);
		delete m_world;
	}
	m_world = w;
	m_resourceDock->setWorld(m_world);
	m_regionMapWidget->setWorld(m_world);
	m_railMan->setWorld(m_world);
	m_mapMapWidget->setWorld(m_world);
	m_fileMan->setWorld(m_world);
	m_consoleMan->setWorld(m_world);
	setWindowTitle(QString("%1 - yamce").arg(m_world->getName()));
	int rcount = m_world->getGroup(wgOverworldRegions)->getFiles().size();
	int mcount = m_world->getGroup(wgMaps)->getFiles().size();
	QString s = QString("%1: %2 regions, %3 maps").arg(m_world->getName()).arg(rcount).arg(mcount);
	m_statusBar->showMessage(s);
	setupWidgets();
	qInfo() << "loadWorld:" << m_world->getName();
	return(true);
}

void MainWindow::closeEvent(QCloseEvent*)
{
	if (m_world) {
		captureWidgets();
		m_world->m_yamceDoc.save(this, m_world);
	}
	m_resourceDock->closing();
	m_mapWidget->closing();
}

static QString s_MainWindow			("MainWindow");
static QString s_geometry			("geometry");
static QString s_state				("state");
static QString s_directory			("directory");
static QString s_helpDirectory		("helpDirectory");
static QString s_showDebug			("showDebug");

void MainWindow::savePreferences()
{
	QSettings settings(QSettings::IniFormat, QSettings::UserScope, s_companyName, s_productName);
	settings.beginGroup(s_MainWindow);
	settings.setValue(s_geometry, saveGeometry());
	settings.setValue(s_state, saveState());
	settings.setValue(s_directory, m_currentDirectory);
	settings.setValue(s_helpDirectory, m_helpMan->getHelpDirectory());
	settings.setValue(s_showDebug, m_consoleMan->getShowDebug());
	settings.endGroup();
}

void MainWindow::loadPreferences()
{
	QSettings settings(QSettings::IniFormat, QSettings::UserScope, s_companyName, s_productName);
	QVariant qv;
	settings.beginGroup(s_MainWindow);
	this->restoreGeometry(settings.value(s_geometry).toByteArray());
	this->restoreState(settings.value(s_state).toByteArray());
	qv = settings.value(s_directory);
	this->m_currentDirectory = qv.toString();
	qv = settings.value(s_helpDirectory);
	if (qv.isValid())
		this->m_helpMan->setHelpDirectory(qv.toString());
	qv = settings.value(s_showDebug);
	if (qv.isValid())
		this->m_consoleMan->setShowDebug(qv.toBool());
}

void MainWindow::captureWidgets()
{
	m_mapWidget->captureWidgets();
	m_consoleMan->updateConsoleDialog();
	m_resourceDock->captureWidgets();
	this->capturePrefsDialog();
}

void MainWindow::resetWidgets()
{
	m_mapWidget->resetMapSettingsDialog();
	m_mapWidget->resetRailManDialog();
	m_consoleMan->updateConsoleDialog();
	m_mainToolbar->resetWidgets();

}

void MainWindow::setupWidgets()
{
	YamceDoc& doc = m_world->m_yamceDoc;
	m_mapWidget->setupWidgets(&doc);
	m_consoleMan->setupConsoleDialog(&doc.m_consoleDialogSettings);

	m_resourceDock->setupColumnWidths();
	bool doEmit = m_resourceDock->setGroupDropdownValue(doc.m_currentResourceGroup);
	if (doEmit) {
		emit(changeCurrentResourceGroup(doc.m_currentResourceGroup));
		resourceDropdownChanged(doc.m_currentResourceGroup);
	}
	m_mainToolbar->setZoomValue(doc.m_currentZoom);
	if (doc.m_prefsDialogSettings.m_shown)
		this->onPreferences();

}

void MainWindow::setResourceColumnWidths(int* widths, int numWidths)
{
	m_world->m_yamceDoc.m_resourceColumnWidths.clear();
	for (int i=0; i<numWidths; i++) {
		m_world->m_yamceDoc.m_resourceColumnWidths.append(QVariant(*widths));
		widths++;
	}
}

void MainWindow::getResourceColumnWidths(int* widths, int numWidths)
{
	for (int i=0; i<numWidths; i++) {
		int n = 0;
		if (i < m_world->m_yamceDoc.m_resourceColumnWidths.size())
			n = m_world->m_yamceDoc.m_resourceColumnWidths.at(i).toInt();
		*widths++ = n;
	}
}

void MainWindow::saveWorld()
{
	if (DEBUG)
		qInfo() << "Saving world" << m_world->getName();
	if (m_world) {
		captureWidgets();
		m_world->save();
		m_commandMan->worldSaved();
		qInfo() << "World saved";
	}
}

void MainWindow::resourceDropdownChanged(int newSet)
{
	if (DEBUG)
		qDebug() << "resourceDropdownChanged" << newSet;
	m_world->m_yamceDoc.m_currentResourceGroup = newSet;
	emit(changeCurrentResourceGroup(newSet));
	switch (newSet) {
	case wgVillages:
	case wgPlayers:
		foreach(NBTFile* file, m_world->getGroup(newSet)->getFiles()) {
			if (!file->isLoaded())
				m_fileMan->readFile(file);
		}
		break;
	case wgMaps:
		int total = m_world->getGroup(wgMaps)->getFiles().size();
		m_statusBar->setProgressText("Loaded %1 of %2 maps", total);
		foreach(NBTFile* file, m_world->getGroup(wgMaps)->getFiles()) {
			MapFile* mi = (MapFile*)file;
			if (!mi->hasMapImage() && !mi->isLoaded())
				m_fileMan->readFile(mi);
		}
		break;
	}
}

void MainWindow::setProgress(ProgressSet* set)
{
	m_statusBar->setProgress(set);
}

void MainWindow::onZoomValueChanged(double value)
{
	m_world->m_yamceDoc.m_currentZoom = value;
	emit(changeZoomValue(value));
}

void MainWindow::regionChanged(int x, int z)
{
	RegionFile* file = m_world->findRegion(x, z);
	if (!file) {
		qInfo() << "MainWindow:regionChanged Ignoring not found region" << x << "/" << z;
		return;
	}
	if (DEBUG)
		qDebug() << "MainWindow:regionChanged" << x << "/" << z;
	m_resourceDock->setRegionFile(file);
	if (!file->isLoaded()) {
		m_fileMan->readFile(file);
	}
}

void MainWindow::fileLoaded(NBTFile* file)
{
	MapFile* mf = dynamic_cast<MapFile*>(file);
	if (mf) {
		m_statusBar->incProgressDone();
	}
}

void MainWindow::exportPNG()
{
//	if (m_world->m_yamceDoc.m_exportPngName.isEmpty()) {
//		QString s = QFileDialog::getSaveFileName(this, tr("Choose name of MapWall PNG file"));
//		if (s.isEmpty()) {
//			qInfo() << "exportPng cancelled";
//			return;
//		}
//		m_world->m_yamceDoc.m_exportPngName = s;
//	}
	qInfo() << "Exporting MapWall PNG to" << m_world->getMapWallFileName();
	emit(doExportPNG());
}

void MainWindow::errorMsgReceived(QString msg)
{
	errorMsgHandler(msg);
}

void MainWindow::errorMsgHandler(const QString& msg)
{
	QMessageBox::critical(this, "Error", msg, QMessageBox::NoButton);
}

void MainWindow::onPreferences()
{
	if (!m_prefsDialog) {
		m_prefsDialog = new PreferencesDialog(this);
		connect(m_prefsDialog, SIGNAL(rejected()), this, SLOT(onPreferencesClosed()));
		connect(m_prefsDialog, SIGNAL(accepted()), this, SLOT(onPreferencesClosed()));
		if (!m_world->m_yamceDoc.m_prefsDialogSettings.m_pos.isNull()) {
			m_prefsDialog->move(m_world->m_yamceDoc.m_prefsDialogSettings.m_pos.topLeft());
			m_prefsDialog->resize(m_world->m_yamceDoc.m_prefsDialogSettings.m_pos.size());
			m_world->m_yamceDoc.m_prefsDialogSettings.m_shown = true;
		}
	}
	m_prefsDialog->show();
	m_prefsDialog->raise();
}

void MainWindow::onPreferencesClosed()
{
	capturePrefsDialog();
	m_prefsDialog = nullptr;
	m_world->m_yamceDoc.m_prefsDialogSettings.m_shown = false;
}

void MainWindow::capturePrefsDialog()
{
	if (!m_world)
		return;
	WindowSettings* ws = &m_world->m_yamceDoc.m_prefsDialogSettings;
	if (m_prefsDialog) {
		ws->m_pos.setTopLeft(m_prefsDialog->pos());
		ws->m_pos.setSize(m_prefsDialog->size());
	}
	ws->m_shown = (m_prefsDialog != nullptr);
}

YamceDoc* MainWindow::getYamceDoc()
{
	if (m_world)
		return(&m_world->m_yamceDoc);
	return(nullptr);
}
