/******************************************************************************
 * mainwindow.h - The standard Qt MainWindow
 *
 * yamce - Yet Another Minecraft Editor
 * Copyright(c) 2018 - Dick Balaska, and BuckoSoft.
 *
 * yamce is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * yamce is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/
#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include <QMainWindow>

#include "errormsgreceiver.h"
#include "syncman.h"

class CentralWidget;
class CommandMan;
class ConsoleMan;
class FileMan;
class HelpMan;
class MainToolbar;
class MapMapWidget;
class MapWidget;
class NBTFile;
class PreferencesDialog;
class ProgressSet;
class RailMan;
class ResourceDock;
class RegionMapWidget;
class RegionMapScrollArea;
class StatusBar;
class SyncMan;
class World;
class YamceDoc;

class MainWindow : public QMainWindow, public ErrorMsgReceiver
{
	Q_OBJECT

public:
	MainWindow(QWidget *parent = 0);
	~MainWindow();

	void setResourceColumnWidths(int* widths, int numWidths);
	void getResourceColumnWidths(int* widths, int numWidths);

	CommandMan*		getCommandMan();
	ConsoleMan*		getConsoleMan();
	FileMan*		getFileMan();
	HelpMan*		getHelpMan();
	MainToolbar*	getMainToolbar();
	MapMapWidget*	getMapMapWidget();
	MapWidget*		getMapWidget();
	RailMan*		getRailMan();
	ResourceDock*	getResourceDock();
	StatusBar*		getStatusBar();
	SyncMan*		getSyncMan();
	bool			isSyncManActive();
	World*			getWorld();
	YamceDoc*		getYamceDoc();

	void	setProgress(ProgressSet* set);
	void	incProgress(ProgressSet* set);
signals:
	void	changeZoomValue(double value);
	void	changeCurrentResourceGroup(int newGroup);
	void	doExportPNG();

public slots:
	void	openWorld();
	bool	loadWorld();
	void	saveWorld();
	void	resourceDropdownChanged(int newSet);
	void	onZoomValueChanged(double value);
	void	regionChanged(int x, int z);
	void	fileLoaded(NBTFile* file);
	void	exportPNG();
	void	errorMsgReceived(QString msg);
	void	onPreferences();

private slots:
	void	onPreferencesClosed();

protected:
	virtual	void closeEvent(QCloseEvent*);
	virtual void errorMsgHandler(const QString& msg);

protected:
	void	loadPreferences();
	void	savePreferences();

private:
	void	captureWidgets();
	void	resetWidgets();
	void	setupWidgets();
	void	capturePrefsDialog();


	ResourceDock*		m_resourceDock;
	MainToolbar*		m_mainToolbar;
	StatusBar*			m_statusBar;
	CentralWidget*		m_centralWidget;
	MapMapWidget*		m_mapMapWidget;
	MapWidget*			m_mapWidget;
	RegionMapWidget*	m_regionMapWidget;
	RegionMapScrollArea*	m_regionMapScrollArea;
	QString				m_currentDirectory;
	World*				m_world;
	FileMan*			m_fileMan;
	RailMan*			m_railMan;
	CommandMan*			m_commandMan;
	ConsoleMan*			m_consoleMan;
	SyncMan*			m_syncMan;
	PreferencesDialog*	m_prefsDialog;
	HelpMan*			m_helpMan;

};

inline FileMan*			MainWindow::getFileMan() { return(m_fileMan); }
inline CommandMan*		MainWindow::getCommandMan() { return(m_commandMan); }
inline ConsoleMan*		MainWindow::getConsoleMan() { return(m_consoleMan); }
inline HelpMan*			MainWindow::getHelpMan() { return(m_helpMan); }
inline MapMapWidget*	MainWindow::getMapMapWidget() { return(m_mapMapWidget); }
inline MapWidget*		MainWindow::getMapWidget() { return(m_mapWidget); }
inline MainToolbar*		MainWindow::getMainToolbar() { return(m_mainToolbar); }
inline RailMan*			MainWindow::getRailMan() { return(m_railMan); }
inline ResourceDock*	MainWindow::getResourceDock() { return(m_resourceDock); }
inline StatusBar*		MainWindow::getStatusBar() { return(m_statusBar); }
inline World*			MainWindow::getWorld() { return(m_world); }
inline SyncMan*			MainWindow::getSyncMan() { return(m_syncMan); }
inline bool				MainWindow::isSyncManActive() { return(m_syncMan->isActive()); }
#endif // MAINWINDOW_H
