/******************************************************************************
 * mapwidget.cpp - The central widget that displays the map of the maps
 *
 * yamce - Yet Another Minecraft Editor
 * Copyright(c) 2018 - Dick Balaska, and BuckoSoft.
 *
 * yamce is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * yamce is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/
#include <QDebug>
#include <QVBoxLayout>
#include <QResizeEvent>
#include <QComboBox>
#include <QLabel>
#include <QPainter>
#include <QCheckBox>
#include <QPushButton>
#include <QScroller>
#include <math.h>

#include "nbtobject.h"

#include "mapfile.h"
#include "mapbannerpainter.h"
#include "mapsettingsdialog.h"
#include "railmandialog.h"
#include "mapwidget.h"
#include "fileman.h"
#include "railman.h"
#include "playerfile.h"
#include "mainwindow.h"
#include "yamcedoc.h"

static bool DEBUG = false;
enum {
	mz0,
	mz1,
	mz2,
	mz3,
	mz4,
	mzExplorer,
	mzAll
};

MapWidget::MapWidget(MainWindow* mainWindow, QWidget* parent)
	: QWidget(parent),
	  m_mainWindow(mainWindow),
	  m_mapSettingsDialog(nullptr),
	  m_railManDialog(nullptr),
	  m_mapSettingsDialogSettings(nullptr),
	  m_railManDialogSettings(nullptr)
{
	QVBoxLayout* layout = new QVBoxLayout();
	QHBoxLayout* hlayout = new QHBoxLayout();
	//-----
	QLabel* l = new QLabel();
	l->setText("Zoom:");
	hlayout->addWidget(l);
	//-----
	m_mapZoomDisplayed = new QComboBox();
	hlayout->addWidget(m_mapZoomDisplayed);
	hlayout->addStretch(2);
	QStringList sl;
	sl << "1 (128x128)" << "2 (256x256)" << "4 (512x512)" << "8 (1024x1024)" << "16 (2048x2048)" << "Explorer maps" << "All";
	m_mapZoomDisplayed->addItems(sl);
	//-----
	m_displayMapImages = new QCheckBox();
	m_displayMapImages->setText("&Images");
	m_displayMapImages->setChecked(true);
	m_displayMapImages->setShortcut(QKeySequence(Qt::ALT + Qt::Key_I));
	hlayout->addWidget(m_displayMapImages);

	//-----
	m_settingsButton = new QPushButton(QIcon(":resources/icons/map_24x24.png"), "");
	connect(m_settingsButton, SIGNAL(clicked(bool)), this, SLOT(onMapSettingsDialog(bool)));
	hlayout->addWidget(m_settingsButton);
	m_railsButton = new QPushButton(QIcon(":resources/icons/rail_24x24.png"), "");
	connect(m_railsButton, SIGNAL(clicked(bool)), this, SLOT(onRailManDialog(bool)));
	hlayout->addWidget(m_railsButton);
	hlayout->addStretch(100);

	//-----
	layout->addLayout(hlayout);
	m_mapMapScrollArea = new MapMapScrollArea(mainWindow);
	QScroller::grabGesture(m_mapMapScrollArea, QScroller::LeftMouseButtonGesture);
	connect(m_displayMapImages, SIGNAL(stateChanged(int)), m_mapMapScrollArea->m_widget, SLOT(displayImagesChanged(int)));
	connect(m_mapZoomDisplayed, SIGNAL(currentIndexChanged(int)), m_mapMapScrollArea->m_widget, SLOT(zoomsDisplayedChanged(int)));
	m_mapMapWidget = m_mapMapScrollArea->m_widget;
	m_mapMapWidget->zoomsDisplayedChanged(mz0);
	layout->addWidget(m_mapMapScrollArea);

	this->setLayout(layout);
}

MapWidget::~MapWidget()
{
	if (m_mapSettingsDialog) {
		m_mapSettingsDialogSettings = nullptr;
		m_mapSettingsDialog->close();
	}
}

void MapWidget::closing()
{
	if (m_railManDialog) {
		m_railManDialog->closing();
		m_railManDialog->close();
	}
}

void MapWidget::setupWidgets(YamceDoc* doc)
{
	m_doc = doc;
	setupMapSettingsDialog(&doc->m_mapSettingsDialogSettings);
	setupRailManDialog(&doc->m_railManDialogSettings);
	setMapsImagesDisplayed(doc->m_mapSettings.m_displayMapImages);
	setMapsZoomDisplayed(doc->m_mapSettings.m_mapZoomDisplayed);
	m_mapMapWidget->mapWallChanged();

}
void MapWidget::captureWidgets()
{
	updateMapSettingsDialog();
	updateRailManDialog();
	m_doc->m_mapSettings.m_displayMapImages = m_displayMapImages->isChecked();
	m_doc->m_mapSettings.m_mapZoomDisplayed = m_mapZoomDisplayed->currentIndex();
}

void MapWidget::setMapsZoomDisplayed(int index)
{
	m_mapZoomDisplayed->setCurrentIndex(index);
}

void MapWidget::setMapsImagesDisplayed(bool b)
{
	m_displayMapImages->setChecked(b);
	m_mapMapWidget->m_displayImages = b;
}

void MapWidget::onMapSettingsDialog(bool)
{
	if (!m_mapSettingsDialog) {
		m_mapSettingsDialog = new MapSettingsDialog(this, m_mainWindow);
		m_mapSettingsDialog->setWorld(m_mapMapWidget->m_world);
		connect(m_mapSettingsDialog, SIGNAL(rejected()), this, SLOT(onMapSettingsDialogClosed()));
		connect(m_mapSettingsDialog, SIGNAL(accepted()), this, SLOT(onMapSettingsDialogClosed()));
		connect(m_mapSettingsDialog, SIGNAL(mapSettingsChanged()), this, SLOT(onMapSettingsChanged()));
		if (m_mapSettingsDialogSettings) {
			if (!m_mapSettingsDialogSettings->m_pos.isNull()) {
				m_mapSettingsDialog->move(m_mapSettingsDialogSettings->m_pos.topLeft());
				m_mapSettingsDialog->resize(m_mapSettingsDialogSettings->m_pos.size());
			}
			m_mapSettingsDialogSettings->m_shown = true;
		}
	}
	m_mapSettingsDialog->show();
	m_mapSettingsDialog->raise();

}

void MapWidget::onRailManDialog(bool)
{
	if (!m_railManDialog) {
		m_railManDialog = new RailManDialog(this, m_mainWindow);
		m_railManDialog->setWorld(m_mapMapWidget->m_world);
		connect(m_railManDialog, SIGNAL(rejected()), this, SLOT(onRailManDialogClosed()));
		connect(m_railManDialog, SIGNAL(accepted()), this, SLOT(onRailManDialogClosed()));
		connect(m_railManDialog, SIGNAL(mapSettingsChanged()), this, SLOT(onRailManDone()));
		connect(m_railManDialog, SIGNAL(processRailsChecked()), this, SLOT(onProcessRailsChecked()));
		connect(m_mainWindow->getFileMan(), SIGNAL(readFileDone(NBTFile*)), m_railManDialog, SLOT(fileLoaded(NBTFile*)));
		if (m_railManDialogSettings) {
			if (!m_railManDialogSettings->m_pos.isNull()) {
				m_railManDialog->move(m_railManDialogSettings->m_pos.topLeft());
				m_railManDialog->resize(m_railManDialogSettings->m_pos.size());
				m_railManDialogSettings->m_shown = true;
			}
		}
	}
	m_railManDialog->show();
	m_railManDialog->raise();

}

void MapWidget::setupMapSettingsDialog(WindowSettings* settings)
{
	m_mapSettingsDialogSettings = settings;
	if (settings->m_shown) {
		onMapSettingsDialog(true);
	}
}

void MapWidget::updateMapSettingsDialog()
{
	if (m_mapSettingsDialogSettings) {
		if (m_mapSettingsDialog) {
			m_mapSettingsDialogSettings->m_pos.setTopLeft(m_mapSettingsDialog->pos());
			m_mapSettingsDialogSettings->m_pos.setSize(m_mapSettingsDialog->size());
		}
		m_mapSettingsDialogSettings->m_shown = (m_mapSettingsDialog != nullptr);
	}
}

void  MapWidget::resetMapSettingsDialog()
{
	m_mapSettingsDialogSettings = nullptr;
	if (m_mapSettingsDialog)
		m_mapSettingsDialog->close();
}

void MapWidget::setupRailManDialog(WindowSettings* settings)
{
	m_railManDialogSettings = settings;
	if (settings->m_shown) {
		onRailManDialog(true);
	}
}

void MapWidget::updateRailManDialog()
{
	if (m_railManDialogSettings) {
		if (m_railManDialog) {
			m_railManDialogSettings->m_pos.setTopLeft(m_railManDialog->pos());
			m_railManDialogSettings->m_pos.setSize(m_railManDialog->size());
		}
		m_railManDialogSettings->m_shown = (m_railManDialog != nullptr);
	}
}

void  MapWidget::resetRailManDialog()
{
	m_railManDialogSettings = nullptr;
	if (m_railManDialog)
		m_railManDialog->close();
}

void MapWidget::onMapSettingsDialogClosed()
{
	updateMapSettingsDialog();
	m_mapSettingsDialog = nullptr;
	if (m_mapSettingsDialogSettings)
		m_mapSettingsDialogSettings->m_shown = false;
}

void MapWidget::onRailManDialogClosed()
{
	updateRailManDialog();
	m_railManDialog = nullptr;
	if (m_railManDialogSettings)
		m_railManDialogSettings->m_shown = false;
}

void MapWidget::onMapSettingsChanged()
{
	m_mapMapWidget->mapWallChanged();
}

void MapWidget::onProcessRailsChecked()
{
	emit(processRailsChecked());
}

void MapWidget::onRailManDone()
{
	if (m_railManDialog)
		m_railManDialog->onRailManDone();
	m_mapMapWidget->onRailManDone();
}
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
MapMapScrollArea::MapMapScrollArea(MainWindow* mainWindow, QWidget* parent)
	: QScrollArea(parent)
{
	setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOn);
	m_widget = new MapMapWidget(mainWindow, this);
	this->setWidget(m_widget);
}

void MapMapScrollArea::resizeEvent(QResizeEvent *event)
{
	if (DEBUG)
		qDebug() << "MapMapScrollArea:resizeEvent" << event->size();
	super::resizeEvent(event);
	//m_widget->resize(event->size()*m_widget->m_zoom);
}

void MapMapScrollArea::zoomValueChanged(double value)
{
	m_widget->m_zoom = value;
	m_widget->resize(this->size()*m_widget->m_zoom);
}


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
MapMapWidget::MapMapWidget(MainWindow* mainWindow, QWidget* parent)
	: QWidget(parent),
	  m_parent(parent),
	  m_mainWindow(mainWindow),
	  m_zoom(1.0),
	  m_maps(nullptr)
{
	resetBounds();
}

MapMapWidget::~MapMapWidget()
{
	while (!m_mapSummaries.isEmpty()) {
		delete m_mapSummaries.takeLast();
	}
}

void MapMapWidget::setWorld(World *world)
{
	m_world = world;
	m_maps = world->getGroup(wgMaps);
	m_mapSettings = &world->m_yamceDoc.m_mapSettings;
	m_zoom = world->m_yamceDoc.m_currentZoom;
	resetBounds();
	while (!m_mapSummaries.isEmpty())
		delete m_mapSummaries.takeFirst();
	mapWallChanged();
}


void MapMapWidget::resetBounds()
{
	m_bounds = QRect();		// empty
}

void MapMapWidget::paintEvent(QPaintEvent*)
{
	QPainter p(this);
	paintMapWall(p);
}

void MapMapWidget::paintMapWall(QPainter& p)
{
	if (DEBUG)
		qDebug() << "MapMapWidget:paintMapWall ********************************************";
	if (DEBUG)
		qDebug() << "MapMapWidget:paint bounds=" << m_bounds;
	QRect rect;
	if (m_displayImages) {
		foreach (NBTFile* file, m_maps->getFiles()) {
			MapFile* mf = (MapFile*)file;
			if (!mf->m_pixmap)
				continue;
			if (mf->m_dimension)
				continue;
			if (mf->m_imageProcessed && mf->m_imageDisplayed) {
				if (m_zoomsDisplayed != mzAll) {
					if (m_zoomsDisplayed == mzExplorer && !mf->isExplorerMap())
						continue;
					if (m_zoomsDisplayed != mf->m_scale)
						continue;
				}
				determineMapRect(mf->m_pos, mf->m_scale, rect);
				rect.moveTopLeft(rect.topLeft()-m_bounds.topLeft());
				if (DEBUG)
					qDebug() << "paint" << rect << mf->getFileName();
				p.drawPixmap(rect, *mf->getPixmap());
				foreach (MapBanner* mb, mf->m_mapBanners) {
					QPoint pos = determineBannerPos(mf->m_scale, mb->m_pos);
					if (DEBUG)
						qDebug() << "paintBanner" << pos << mb->m_color;
					MapBannerPainter::draw(p, pos, mb->m_color, m_zoom);
				}
			}
		}
		if (m_mapSettings->m_drawGridLines) {
			QPen pen(Qt::black);
			pen.setWidth(m_mapSettings->m_gridWidth);
			p.setPen(pen);
			//qDebug() << "draw grid" << width() << height();
			double	gridDist = 128.0 * m_zoom;
			for (double col=0; col <= width()+1; col+=gridDist) {
				p.drawLine(col, 0, col, height());
			}
			for (double row=0; row <= height()+1; row+=gridDist) {
				p.drawLine(0, row, width(), row);
			}
		}
	} else {
		QRect rect;
		QFont font = p.font();
		QFontMetrics fm(font);
		m_fontHeight = fm.height();
		QPoint pt;
		foreach (MapSummary* ms, m_mapSummaries) {
			int row=1;
			bool boxDrawn = false;
			foreach(MapFile* mf, ms->m_files) {
				if (m_zoomsDisplayed != mzAll) {
					if (m_zoomsDisplayed == mzExplorer && !mf->isExplorerMap())
						continue;
					if (m_zoomsDisplayed != mf->m_scale)
						continue;
				}
				determineMapRect(mf->m_pos, mf->m_scale, rect);
				rect.moveTopLeft(rect.topLeft()-m_bounds.topLeft());
				rect.setWidth(128*m_zoom-1);
				rect.setHeight(128*m_zoom-1);
				if (!boxDrawn) {
					p.setPen(Qt::black);
					p.drawRect(rect);
					if (DEBUG)
						qDebug() << "MapMapWidget:paintEvent rect=" << rect;
					boxDrawn = true;
				}
				pt.setX(rect.left());
				pt.setY(rect.top()+m_fontHeight*row);
				if (mf->m_imageDisplayed)
					p.setPen(Qt::black);
				else
					p.setPen(Qt::red);
				p.drawText(pt, mf->getFileName());
#if 1		// mouse click debug
				QRect frect;
				frect.setWidth(128*m_zoom-1);
				frect.setHeight(m_fontHeight);
				frect.moveLeft(rect.left());
				frect.moveTop(m_fontHeight*(row-1)+rect.top());
				p.drawRect(frect);
#endif
				if (DEBUG)
					qDebug() << "MapMapWidget:paintEvent frect=" << frect << mf->getFileName();
				row++;
			}
		}
	}
	if (m_mainWindow->isSyncManActive()) {
		double lzoom = m_zoom;
		if (lzoom < 1.0)
			lzoom = 1.0;
		QVector<QLineF>	lines;
		double funf = 5.0*lzoom;
		double acht = 8.0*lzoom;
		double zehn = 10.0*lzoom;
		lines.append(QLineF(-funf, -funf, funf, -funf));
		lines.append(QLineF(-funf, -funf, 0,  acht));
		lines.append(QLineF( funf, -funf, 0,  acht));

		QVector<QPointF> points;
		points.append(QPointF(-funf, -funf));
		points.append(QPointF( 0.0,  acht));
		points.append(QPointF( funf, -funf));
		QPolygonF polygon(points);
		QPainterPath	playerPath;
		playerPath.addPolygon(polygon);

		QPainterPath	playerOffScreenPath;
		playerOffScreenPath.addEllipse(-funf, -funf, zehn, zehn);
		QPoint boundsCorner;
		QPen pen(Qt::white);
		pen.setWidth(2);
		QBrush brush(Qt::white);
		int scaler = 1 << m_zoomsDisplayed;
		//QPoint c(64*scaler, 64*scaler);
		boundsCorner = m_bounds.topLeft()*scaler;
		Group* wg = m_world->getGroup(wgPlayers);
		for (NBTFile* file : wg->getFiles()) {
			if (!file->isLoaded())
				continue;
			PlayerFile* pf = static_cast<PlayerFile*>(file);
			QPoint pt = pf->getPos();
			//qDebug() << "paintMap: playerPos=" << pt;
			int half = 64.0*m_zoom;
			pt = ((pt*m_zoom)-boundsCorner) / scaler + QPoint(half, half);
			//pt *= m_zoom;
			QTransform transform;
			p.setPen(pen);
			if (pt.x() < 0 || pt.y() < 0 || pt.x() > width() || pt.y() > height()) {
				if (pt.x() < 0)
					pt.setX(4);
				if (pt.y() < 0)
					pt.setY(4);
				if (pt.x() > width())
					pt.setX(width()-4);
				if (pt.y() > height())
					pt.setY(height()-4);
				transform.translate(pt.x(), pt.y());
				transform.rotate(pf->getRot());
				p.setTransform(transform);
				p.fillPath(playerOffScreenPath, brush);
				transform.scale(1.4*m_zoom, 1.4*m_zoom);
				p.setPen(Qt::red);
				p.drawPath(playerOffScreenPath);
			} else {
				transform.translate(pt.x(), pt.y());
				transform.rotate(pf->getRot());
				p.setTransform(transform);
				transform.scale(m_zoom, m_zoom);
				p.fillPath(playerPath, brush);
				transform.scale(1.4*m_zoom, 1.4*m_zoom);
				p.setPen(Qt::black);
				p.drawLines(lines);
			}
		}
	}
}

void MapMapWidget::mousePressEvent(QMouseEvent* event)
{
	MapFile* foundMap = nullptr;
	QRect frect;
	frect.setWidth(128);
	frect.setHeight(m_fontHeight);
	qDebug() << "MapMapWidget:mousePress" << "mousePos=" << event->pos();

	foreach (MapSummary* ms, m_mapSummaries) {
		int row=1;
		foreach (MapFile* mf, ms->m_files) {
			if (m_zoomsDisplayed != mzAll) {
				if (m_zoomsDisplayed == mzExplorer && !mf->isExplorerMap())
					continue;
				if (m_zoomsDisplayed != mf->m_scale)
					continue;
			}
			determineMapRect(mf->m_pos, mf->m_scale, frect);
			if (m_displayImages) {
				frect.moveTopLeft(frect.topLeft()-m_bounds.topLeft());
				if (mf->m_imageDisplayed && frect.contains(event->pos())) {
					foundMap = mf;		// save the map, we want the bottom one in the list (the one actually displayed)
				}
			} else {
				frect.moveTopLeft(frect.topLeft()-m_bounds.topLeft());
				frect.moveTop(m_fontHeight*(row-1)+frect.top());
				frect.setWidth(128);
				frect.setHeight(m_fontHeight);
				if (DEBUG)
					qDebug() << "MapMapWidget:mousePress" << "frect=" << frect << mf->getFileName();
				if (frect.contains(event->pos())) {
					qDebug() << "MapMapWidget:mousePress" << "got" << mf->getFileName();

					mf->m_imageDisplayed = !mf->m_imageDisplayed;
					repaint();
					return;
				}
			}
			row++;
		}
		if (m_displayImages && foundMap) {
			emit(mapSelected(foundMap));
			return;
		}
	}
}

void MapMapWidget::zoomsDisplayedChanged(int level)
{
	m_zoomsDisplayed = level;
	mapWallChanged();
}

void MapMapWidget::onZoomChanged(double value)
{
	m_zoom = value;
	mapWallChanged();
}

void MapMapWidget::mapWallChanged(bool doRepaint)
{
	if (!m_maps)
		return;
	QRect oldBounds = m_bounds;
	resetBounds();
	foreach(NBTFile* file, m_maps->getFiles()) {
		MapFile* mf = (MapFile*)file;
		if (mf->hasMapImage() && mf->isImageDisplayed()) {
			if (m_zoomsDisplayed != mzAll) {
				if (m_zoomsDisplayed == mzExplorer && !mf->isExplorerMap())
					continue;
				if (m_zoomsDisplayed != mf->m_scale)
					continue;
			}
			bool resized = updateBoundsRect(mf);
			if (DEBUG)
				qDebug() << "MapMapWidget:resizeMapWall" << file->getFileName() << "bounds:" << m_bounds << "resized:" << resized;
		}
	}
	if (DEBUG)
		qDebug() << "MapMapWidget:mapWallChanged size=" << m_bounds.size() << "parent=" << m_parent->size();
	if (m_bounds == oldBounds && doRepaint) {
		this->repaint();
	} else {
		this->resize(m_bounds.size());
		if (m_bounds.size().width() < m_parent->size().width()
				&& m_bounds.size().height() < m_parent->size().height()
				&& doRepaint) {
			this->repaint();
		}
	}
}

QPoint MapMapWidget::determineBannerPos(int mapScale, const QPoint& bannerPos)
{
	int scaler = 1 << (mapScale);
	QPoint boundsCorner;
	boundsCorner = m_bounds.topLeft()*scaler;
	QPoint pt = bannerPos;
	int zh = 64.0*m_zoom;
	pt = ((pt*m_zoom)-boundsCorner) / scaler + QPoint(zh, zh);

	return(pt);
}

bool MapMapWidget::determineMapRect(const QPoint point, int scale, QRect& rectr)
{
	QRectF rectf;
	double scaler = (double)((int)(1 << scale));
	rectf.setLeft(point.x()/scaler);
	rectf.setTop(point.y()/scaler);

	if (m_zoomsDisplayed == mzAll)
		rectf.setWidth(128*scaler);
	else
		rectf.setWidth(128);
	rectf.setHeight(rectf.width());
	rectf.setRect(rectf.x()*m_zoom, rectf.y()*m_zoom, rectf.width()*m_zoom, rectf.height()*m_zoom);
	if (m_mapSettings->m_drawGridLines && m_mapSettings->m_expandForGrid) {
		double xofs = point.x() / 128.0;
		rectf.moveLeft(rectf.left()+xofs*m_mapSettings->m_gridWidth);
		rectf.setWidth(rectf.width()+m_mapSettings->m_gridWidth);
		double yofs = point.y() / 128;
		rectf.moveTop(rectf.top()+yofs*m_mapSettings->m_gridWidth);
		rectf.setHeight(rectf.height()+m_mapSettings->m_gridWidth);
		qDebug() << "MapMapWidget:determineMapRect xofs=" << xofs;
	}
	rectr.setRect(floor(rectf.x()), floor(rectf.y()), ceil(rectf.width()), ceil(rectf.height()));
	//	if (DEBUG)
	//qDebug() << "determineMapRect: rect:" << rectf << "rectr:" << rectr;
	return(true);
}

bool MapMapWidget::updateBoundsRect(MapFile* mf)
{
	QRect rect;
	if (m_zoomsDisplayed != mzAll) {
		if (m_zoomsDisplayed == mzExplorer && !mf->isExplorerMap())
			return(false);
		if (m_zoomsDisplayed != mf->m_scale)
			return(false);
	}
	bool res = determineMapRect(mf->m_pos, mf->m_scale, rect);
	if (!res) {
		qCritical() << "MapMapWidget:fileLoaded failed to determine map rect for" << mf->getFileName();
		return(false);
	}
	QRect oldBounds = m_bounds;
	if (m_bounds.isEmpty())
		m_bounds = rect;
	else {
		m_bounds |= rect;
	}

	if (DEBUG)
		qDebug() << "MapMapWidget:updateBoundsRect" << mf->getFileName() << "rect:" << rect;
	return(m_bounds != oldBounds);
}

void MapMapWidget::processMapPosition(MapFile* mf)
{
	if (mf->m_dimension)
		return;
	foreach (MapSummary* ms, m_mapSummaries) {
		if (ms->m_pos == mf->m_pos) {
			foreach (MapFile* mff, ms->m_files) {
				if (mff == mf)
					return;
			}
			ms->m_files.append(mf);
			return;
		}
	}
	MapSummary* ms = new MapSummary();
	ms->m_pos = mf->m_pos;
	ms->m_files.append(mf);
	m_mapSummaries.append(ms);
}

void MapMapWidget::fileLoaded(MapFile* file)
{
	bool resized = updateBoundsRect(file);
	processMapPosition(file);
	if (m_mapSettings->m_drawRailLines)
		applyRailsToMap(file);
	file->m_imageProcessed = true;
	if (DEBUG)
		qDebug() << "MapMapWidget:fileLoaded" << file->getFileName() << "bounds:" << m_bounds << "resized:" << resized;
	if (resized) {
		this->resize(m_bounds.size());
		this->repaint();
	} else {
		this->repaint();
	}
}

void MapMapWidget::fileLoaded(PlayerFile* file)
{
	if (m_mainWindow->isSyncManActive()) {
		if (false)
			qDebug() << "MapMapWidget repaint for player" << file->getFileName() << file->isLoaded();
		repaint();
	}
}

void MapMapWidget::displayImagesChanged(int state)
{
	m_displayImages = state == Qt::Checked;
	repaint();
}

void MapMapWidget::exportPNG()
{
	double oldZoom = m_zoom;
	m_zoom = 1.0;
	mapWallChanged(false);
	QPixmap pm(m_bounds.size());
	pm.fill(Qt::black);
	QPainter p(&pm);
	paintMapWall(p);
	m_zoom = oldZoom;
	mapWallChanged(false);
	QFile file(m_world->getMapWallFileName());
	file.open(QIODevice::WriteOnly);
	pm.save(&file, "PNG");
	file.close();
}

bool MapMapWidget::applyRailsToMap(MapFile *file)
{
	if (!file->m_pixmap)
		return(false);
	bool drawn = false;
	int penWidth = 2;
	const QList<XYZ>& railList = m_mainWindow->getRailMan()->getRails();
	int scaler = 1 << file->m_scale;
	//QPoint halfMap = QPoint(64*scaler, 64*scaler);
	QRect	rect = QRect(file->m_pos.x()-(64*scaler), file->m_pos.y()-(64*scaler), 128*scaler, 128*scaler);

	QPainter p(file->m_pixmap);
	QPen pen;
	if (!m_world)
		pen = QPen(Qt::red);
	else
		pen = QPen(getMapColorMap().at(m_world->m_yamceDoc.m_railColor));
	pen.setWidth(penWidth);
	p.setPen(pen);
	QPoint railPoint;
	QPoint center = QPoint(64, 64);
	foreach(XYZ xyz, railList) {
		if (xyz.inside(rect)) {
			railPoint = QPoint(xyz.x-file->m_pos.x(), xyz.z - file->m_pos.y());
			railPoint /= scaler;
			railPoint += center;
			p.drawPoint(railPoint);
			drawn = true;
		}
	}
	return(drawn);
}

void MapMapWidget::onRailManDone()
{
	if (m_mapSettings->m_drawRailLines) {
		Group* group = m_world->getGroup(wgMaps);
		for (NBTFile* file : group->getFiles()) {
			MapFile* mf = (MapFile*)file;
			applyRailsToMap(mf);
		}
	}
	repaint();
}
