/******************************************************************************
 * mapwidget.h - The central widget that displays the map of the maps
 *
 * yamce - Yet Another Minecraft Editor
 * Copyright(c) 2018 - Dick Balaska, and BuckoSoft.
 *
 * yamce is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * yamce is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/
#ifndef MAPWIDGET_H
#define MAPWIDGET_H

#include <QScrollArea>
#include "world.h"
#include "yamcedoc.h"

class QCheckBox;
class QComboBox;
class QPushButton;

class MainWindow;
class MapFile;
class NBTObject;
class MapSettingsDialog;
class PlayerFile;
class RailManDialog;
class WindowSettings;


class MapMapWidget : public QWidget
{
	Q_OBJECT
public:
	explicit MapMapWidget(MainWindow* mainWindow, QWidget* parent = nullptr);
	virtual ~MapMapWidget();

	//void	setMaps(WGroup* maps);
	void	setWorld(World* world);
	void	fileLoaded(MapFile* file);
	void	fileLoaded(PlayerFile* file);
	void	mapWallChanged(bool doRepaint = true);

signals:
	void	mapSelected(MapFile* mf);

public slots:
	void	zoomsDisplayedChanged(int level);
	void	displayImagesChanged(int state);
	void	exportPNG();
	void	onRailManDone();
	void	onZoomChanged(double value);

protected:
	void paintEvent(QPaintEvent* event) override;
	void mousePressEvent(QMouseEvent* event) override;

private:
	void			paintMapWall(QPainter& p);
	void			resetBounds();
	bool			updateBoundsRect(MapFile* file);
	bool			determineMapRect(const QPoint point, int scale, QRect& rect);
	QPoint			determineBannerPos(int mapScale, const QPoint& bannerPos);
	void			processMapPosition(MapFile* file);
	/// Add the rails data to this map
	/// @return true if any rail data affected this map
	bool			applyRailsToMap(MapFile* file);

	QWidget*			m_parent;
	MainWindow*			m_mainWindow;
	double				m_zoom;
	int					m_zoomsDisplayed;
	bool				m_displayImages;
	World*				m_world;
	Group*				m_maps;
	MapSettings*		m_mapSettings;
	QRect				m_bounds;
	QList<MapSummary*>	m_mapSummaries;
	int					m_fontHeight;
	//MapSettings			m_mapSettings;

	friend class MapWidget;
	friend class MapMapScrollArea;
	friend class MainWindow;
};

class MapMapScrollArea : public QScrollArea
{
private:
	typedef QScrollArea super;
	Q_OBJECT

public:
	MapMapScrollArea(MainWindow* mainWindow, QWidget* parent = nullptr);

	MapMapWidget*	m_widget;

public slots:
	  void	zoomValueChanged(double value);
//	  void	explorerTypeChanged(int newExplorerType);

protected:
	void resizeEvent(QResizeEvent *event) override;
};

class MapWidget : public QWidget
{
private:
	typedef QWidget	super;

	Q_OBJECT

public:
	explicit MapWidget(MainWindow* mainWindow, QWidget* parent = nullptr);
	virtual ~MapWidget();

	MapMapWidget*	getMapMapWidget() { return(m_mapMapScrollArea->m_widget); }
	void			setupWidgets(YamceDoc* doc);
	void			captureWidgets();
	void			resetWidgets();

signals:
	void	processRailsChecked();

public slots:
	void	onMapSettingsDialog(bool b);
	void	onRailManDialog(bool b);
	void	onMapSettingsChanged();
	void	onRailManDone();

private slots:
	void	onMapSettingsDialogClosed();
	void	onRailManDialogClosed();
	void	onProcessRailsChecked();

private:
	void	closing();
	void			setMapsZoomDisplayed(int index);
	void			setMapsImagesDisplayed(bool b);
	void	updateMapSettingsDialog();
	void	resetMapSettingsDialog();
	void	setupMapSettingsDialog(WindowSettings* settings);
	void	updateRailManDialog();
	void	resetRailManDialog();
	void	setupRailManDialog(WindowSettings* settings);

	MainWindow*			m_mainWindow;
	MapMapScrollArea*	m_mapMapScrollArea;
	MapMapWidget*		m_mapMapWidget;
	QComboBox*			m_mapZoomDisplayed;
	QCheckBox*			m_displayMapImages;
	QPushButton*		m_settingsButton;
	QPushButton*		m_railsButton;
	MapSettingsDialog*	m_mapSettingsDialog;
	RailManDialog*		m_railManDialog;
	//QRect				m_mapSettingsDialogPos;
	//QRect				m_railManDialogPos;
	WindowSettings*		m_mapSettingsDialogSettings;
	WindowSettings*		m_railManDialogSettings;
	YamceDoc*			m_doc;

	friend class MainWindow;
};


#endif // MAPWIDGET_H
