/******************************************************************************
 * nbtfile.cpp - Read and write and NBT file
 *
 * yamce - Yet Another Minecraft Editor
 * Copyright(c) 2018 - Dick Balaska, and BuckoSoft.
 *
 * yamce is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * yamce is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/
#include <QFile>
#include <QMessageBox>
#include <QDebug>

#include "mainwindow.h"
#include "qcompressor.h"
#include "nbtobject.h"
#include "nbtfile.h"
#include "errormsgreceiver.h"

static bool DEBUG = false;

static QString s_error	= "Error";


NBTFile::NBTFile()
	: m_object(nullptr),
	  m_loaded(false),
	  m_dirty(false),
	  m_errorMsgReceiver(nullptr),
	  m_fileManReading(false)
{
}

NBTFile::~NBTFile()
{
	 if (m_object)
		 delete m_object;
}

void NBTFile::init(const QString& dirname, const QString& filename, bool compressed)
{
	m_dirName = dirname;
	m_fileName = filename;
	m_compressed = compressed;
}

void NBTFile::unload()
{
	if (m_object) {
		delete m_object;
		m_object = nullptr;
	}
	m_loaded = false;
}

bool NBTFile::load(ErrorMsgReceiver* receiver)
{
	m_errorMsgReceiver = receiver;
	QString path = QString("%1/%2").arg(m_dirName, m_fileName);
	QFile f(path);
	if (!f.exists()) {
		qWarning() << "Can't find file '" << path << "'";
		return(false);
	}
	qint64 size = f.size();
	bool res = f.open(QIODevice::ReadOnly);
	if (!res) {
		QString s = QString("NBTFile failed to open file %1").arg(m_fileName);
		errorMessage(s);
		return(false);
	}
	//QByteArray bf = f.read(size);
	//QByteArray ba;
	char* buff = new char[size];
	qint64 bytesRead = 0;
	while (bytesRead < size) {
		char* p = buff+bytesRead;
		qint64 r = f.read(p, size-bytesRead);
		bytesRead += r;
		if (r == 0)
			break;
		if (r == -1) {
			QString s = QString("NBTFile error reading file %1 got %2 bytes expected %3").arg(m_fileName).arg(bytesRead).arg(size);
			errorMessage(s);
			f.close();
			return(false);
		}
	}
	QByteArray bf(buff, size);
	delete[] buff;

	res = true;
	if (m_compressed)
		res = QCompressor::gzipDecompress(bf, m_data);
	if (!res) {
		QString s = QString("NBTFile::load failed to decompress file %1").arg(m_fileName);
		errorMessage(s);
		return(false);
	}
	//qDebug() << "bf size:" << bf.size();
	if (DEBUG)
		qDebug() << "parsing:" << m_fileName << "size:" << m_data.size();
	m_parseOffset = 0;
	NBTObject* obj = parse(nullptr);
	m_object = obj;
	m_data.clear();
	m_loaded = true;
	f.close();
	return(true);
}


NBTObject*	NBTFile::parse(NBTObject* obj)
{
	forever {
		if (m_parseOffset >= m_data.size()) {
			qWarning() << "parse ran off the end of the file" << m_parseOffset << m_data.size();
			return(obj);
		}
		qint8 type = m_data.at(m_parseOffset++);
		if (type == TAG_End)
			return(obj);
		NBTObject* nobj = new NBTObject;
		nobj->type = type;
		nobj->name = fetchString();
//		if (nobj->name == "Attributes")
//			qDebug() << "nobj->name:" << nobj->name;
		switch (type) {
		case TAG_Byte: {
			qint8 b = m_data.at(m_parseOffset++);
			nobj->d_byte = b;
			break;
		}
		case TAG_Short:	{
			quint16 sh = 0;
			for (int i=0; i<2; i++) {
				sh = (sh << 8) + (uchar)m_data.at(m_parseOffset++);
			}
			nobj->d_short = sh;
			break;
		}
		case TAG_Int: {
			nobj->d_int = fetchInt();
			break;
		}
		case TAG_Long: {
			nobj->d_long = fetchLong();
			break;
		}
		case TAG_String: {
			nobj->d_string = new QString();
			*nobj->d_string = fetchString();
			break;
		}
		case TAG_Double: {
			nobj->d_double = fetchDouble();
			break;
		}
		case TAG_Float: {
			nobj->d_float = fetchFloat();
			break;
		}
		case TAG_Byte_Array: {
			qint32 count = fetchInt();
			nobj->d_byte_array = new QByteArray();
			nobj->d_byte_array->reserve(count);
			for (int i=0; i<count; i++)
				nobj->d_byte_array->append(fetchByte());
			break;
		}
		case TAG_Int_Array: {
			qint32 ncount = fetchInt();
			nobj->size = ncount;
			qint32* p = new qint32[ncount];
			nobj->d_int_array = p;
			for (int i=0; i<ncount; i++) {
				*p = fetchInt();
				p++;
			}
			break;
		}
		case TAG_Long_Array: {
			qint32 ncount = fetchInt();
			nobj->size = ncount;
			qint64* p = new qint64[ncount];
			nobj->d_long_array = p;
			for (int i=0; i<ncount; i++) {
				*p = fetchLong();
				p++;
			}
			break;
		}
		case TAG_List: {
			NList* nlist = new NList();
			nobj->d_list = nlist;
			qint8 ntype = fetchByte();
			qint32 ncount = fetchInt();
			nlist->objtype = ntype;
			nlist->objcount = ncount;
			if (!parseList(nlist))
				return(nullptr);
			break;
		}

		case TAG_Compound:
			parse(nobj);
			break;
		case TAG_End:
			if (obj->type == TAG_Compound) {
				delete nobj;
				return(obj);
			}
			qWarning() << "Unexpected TAG_End @ offset" << m_parseOffset;
			return(nullptr);
		default:
			qCritical() << "Unhandled tag type" << type << "@" << m_parseOffset << "name:" << obj->name;
			return(nullptr);
		}
		if (obj && obj->type == TAG_Compound) {
			if (!obj->d_compound)
				obj->d_compound = new QList<NBTObject*>();
			obj->d_compound->append(nobj);
			continue;
		} else {
			return(nobj);
		}
	}
}

bool NBTFile::parseList(NList* nlist)
{
	int	ncount = nlist->objcount;
	switch (nlist->objtype) {
	case TAG_Compound: {
		nlist->d_compound = new QList<NBTObject*>();
		for (int i=0; i<ncount; i++) {
			NBTObject* cobj = new NBTObject();
			cobj->type = TAG_Compound;
			NBTObject* nobj = parse(cobj);
			if (!nobj) {
				QString e = QString("parse TAG_List error: expected %1 compounds, got %2 @ %3").arg(ncount).arg(i).arg(m_parseOffset);
				errorMessage(e);
				return(false);
			}
			nlist->d_compound->append(cobj);
		}
		break;
	}
	case TAG_List: {
		nlist->d_list = new QList<NList*>();
		for (int i=0; i<ncount; i++) {
			NList* nnlist = new NList();
			nlist->d_list->append(nnlist);
			qint8 nntype = fetchByte();
			qint32 nncount = fetchInt();
			nnlist->objtype = nntype;
			nnlist->objcount = nncount;
			if (!parseList(nnlist))
				return(false);
		}
		break;
	}
	case TAG_Byte: {
		nlist->d_byte = new qint8[ncount];
		qint8* p = nlist->d_byte;
		for (int i=0; i<ncount; i++)
			*p++ = fetchByte();
		break;
	}
	case TAG_Short: {
		nlist->d_short = new qint16[ncount];
		qint16* p = nlist->d_short;
		for (int i=0; i<ncount; i++)
			*p++ = fetchShort();
		break;
	}
	case TAG_Int: {
		nlist->d_int = new qint32[ncount];
		qint32* p = (qint32*)nlist->d_int;
		for (int i=0; i<ncount; i++)
			*p++ = fetchInt();
		break;
	}
	case TAG_String: {
		nlist->d_string = new QStringList();
		for (int i=0; i<ncount; i++)
			nlist->d_string->append(fetchString());
		break;
	}
	case TAG_Float: {
		nlist->d_float = new float[ncount];
		float* p = nlist->d_float;
		for (int i=0; i<ncount; i++)
			*p++ = fetchFloat();
		break;
	}
	case TAG_Double: {
		nlist->d_double = new double[ncount];
		double* p = nlist->d_double;
		for (int i=0; i<ncount; i++)
			*p++ = fetchDouble();
		break;
	}
	case TAG_Int_Array: {
		qint32** pa = new qint32*[ncount];
		nlist->d_int_array = pa;
		for (int ia=0; ia<ncount; ia++) {
			qint32 lcount = fetchInt();
			qint32* p = new qint32[lcount];
			*pa++ = p;
			for (int i=0; i<lcount; i++) {
				*p++ = fetchInt();
			}
		}
		break;
	}
	case TAG_Long_Array: {
		qint64** pa = new qint64*[ncount];
		nlist->d_long_array = pa;
		for (int ia=0; ia<ncount; ia++) {
			qint32 lcount = fetchInt();
			qint64* p = new qint64[lcount];
			*pa++ = p;
			for (int i=0; i<lcount; i++) {
				*p++ = fetchLong();
			}
		}
		break;
	}
	case TAG_End:
		break;
	default: {
		QString e = QString("Unhandled TAG_List type: %1 @ %2").arg(nlist->objtype).arg(m_parseOffset);
		errorMessage(e);
		return(false);
	}
	}
	return(true);
}
qint8 NBTFile::fetchByte()
{
	return(m_data.at(m_parseOffset++));
}

qint16 NBTFile::fetchShort()
{
	quint16 sh = 0;
	for (int i=0; i<2; i++) {
		sh = (sh << 8) + (uchar)m_data.at(m_parseOffset++);
	}
	return(sh);
}

qint32 NBTFile::fetchInt()
{
	quint32 sh = 0;
	for (int i=0; i<4; i++) {
		sh = (sh << 8) + (uchar)m_data.at(m_parseOffset++);
	}
	return(sh);
}

qint64 NBTFile::fetchLong()
{
	quint64 l = 0;
	for (int i=0; i<8; i++) {
		l = (l << 8) + (uchar)m_data.at(m_parseOffset++);
	}
	return(l);
}

QString NBTFile::fetchString()
{
	QString s;
	quint16 size = 0;
	for (int i=0; i<2; i++)
		size = (size << 8) + (uchar)m_data.at(m_parseOffset++);

	for (int i=0; i<size; i++)
		s.append(m_data.at(m_parseOffset++));
	return(s);
}

float NBTFile::fetchFloat()
{
	union {char b[4]; float f; };
	for (int i=0; i<4; i++) {
		quint8 u = m_data.at(m_parseOffset++);
		if (QSysInfo::ByteOrder == QSysInfo::LittleEndian)
			b[3-i] = u;
		else
			b[i] = u;
	}
return(f);
}

double NBTFile::fetchDouble()
{
	union {char b[8]; double d; };
	for (int i=0; i<8; i++) {
		quint8 u = m_data.at(m_parseOffset++);
		if (QSysInfo::ByteOrder == QSysInfo::LittleEndian)
			b[7-i] = u;
		else
			b[i] = u;
	}
return(d);
}


bool NBTFile::save(ErrorMsgReceiver* receiver)
{
	m_errorMsgReceiver = receiver;
	m_parseOffset = 0;
	m_data.clear();
	serializeObject(m_object);
	if (m_compressed) {
		QByteArray ba;
		QCompressor::gzipCompress(m_data, ba);
		m_data = ba;
	}
	QString s = QString("%1/%2").arg(m_dirName, m_fileName);
	QFile file(s);
	if (!file.open(QIODevice::WriteOnly)) {
		QString e = QString("Failed to open for writing file \"%1\"").arg(s);
		errorMessage(e);
		m_data.clear();
		return(false);
	}
	qint64 res = file.write(m_data);
	if (res == -1) {
		QString e = QString("Failed to write file \"%1\"").arg(s);
		errorMessage(e);
		file.close();
		m_data.clear();
		return(false);
	}
	file.close();
	m_data.clear();
	m_dirty = false;
	return(true);
}

void NBTFile::serializeObject(const NBTObject* obj, bool inList)
{
	if (!inList) {
		putByte(obj->type);
		putString(obj->name);
	}
	switch(obj->type) {
	case TAG_Byte:
		putByte(obj->d_byte);
		break;
	case TAG_Short:
		putShort(obj->d_short);
		break;
	case TAG_Int:
		putInt(obj->d_int);
		break;
	case TAG_Long:
		putLong(obj->d_long);
		break;
	case TAG_Float:
		putFloat(obj->d_float);
		break;
	case TAG_Double:
		putDouble(obj->d_double);
		break;
	case TAG_String:
		putString(*obj->d_string);
		break;
	case TAG_Byte_Array:
		putInt(obj->d_byte_array->size());
		for (int i=0; i<obj->d_byte_array->size(); i++)
			putByte(obj->d_byte_array->at(i));
		break;
	case TAG_Compound:
		if (obj->d_compound) {
			foreach(NBTObject* nobj, *obj->d_compound) {
				serializeObject(nobj);
			}
		}
		putByte(TAG_End);
		break;
	case TAG_List: {
		NList* nlist = obj->d_list;
		putByte(nlist->objtype);
		putInt(nlist->objcount);
		switch(nlist->objtype) {
		case TAG_Byte: {
			qint8* p = nlist->d_byte;
			for (int i=0; i<nlist->objcount; i++)
				putByte(*p++);
			break;
		}
		case TAG_Short: {
			qint16* p = nlist->d_short;
			for (int i=0; i<nlist->objcount; i++)
				putShort(*p++);
			break;
		}
		case TAG_Int: {
			qint32* p = nlist->d_int;
			for (int i=0; i<nlist->objcount; i++)
				putInt(*p++);
			break;
		}
		case TAG_Long: {
			qint64* p = nlist->d_long;
			for (int i=0; i<nlist->objcount; i++)
				putLong(*p++);
			break;
		}
		case TAG_Float: {
			float* p = nlist->d_float;
			for (int i=0; i<nlist->objcount; i++)
				putFloat(*p++);
			break;
		}
		case TAG_Double: {
			double* p = nlist->d_double;
			for (int i=0; i<nlist->objcount; i++)
				putDouble(*p++);
			break;
		}
		case TAG_String: {
			QStringList* p = nlist->d_string;
			for (int i=0; i<p->size(); i++)
				putString(p->at(i));
			break;
		}
		case TAG_Compound: {
			foreach(NBTObject* nobj, *nlist->d_compound) {
				serializeObject(nobj, true);
			}
			break;
		}
		case TAG_End:
			break;
		default: {
			QString e = QString("NBTFile:save Serialize data unhandled list type %1").arg(nlist->objtype);
			errorMessage(e);
		}
		}
		break;
	}
	default:
		QString e = QString("NBTFile:save Serialize data unhandled type %1").arg(obj->type);
		errorMessage(e);

	}
}

void NBTFile::putByte(quint8 b)
{
	m_data.append(b);
}

void NBTFile::putShort(qint16 w)
{
	for (int i=1; i>=0; i--)
		m_data.append((w>>(i*8)) & 0xFF);
}
void NBTFile::putInt(qint32 w)
{
	for (int i=3; i>=0; i--)
		m_data.append((w>>(i*8)) & 0xFF);
}
void NBTFile::putLong(qint64 w)
{
	for (int i=7; i>=0; i--)
		m_data.append((w>>(i*8)) & 0xFF);
}

void NBTFile::putFloat(float f)
{
	union {char b[4]; float ff; };
	ff = f;
	for (int i=0; i<4; i++) {
		if (QSysInfo::ByteOrder == QSysInfo::LittleEndian)
			putByte(b[3-i]);
		else
			putByte(b[i]);
	}

}

void NBTFile::putDouble(double d)
{
	union {char b[8]; double dd; };
	dd = d;
	for (int i=0; i<8; i++) {
		if (QSysInfo::ByteOrder == QSysInfo::LittleEndian)
			putByte(b[7-i]);
		else
			putByte(b[i]);
	}

}

void NBTFile::putString(const QString& s)
{
	putShort(s.size());
	m_data.append(s);
}
