/******************************************************************************
 * railman.cpp - Find and Manage the rail locations from the world
 *
 * yamce - Yet Another Minecraft Editor
 * Copyright(c) 2018 - Dick Balaska, and BuckoSoft.
 *
 * yamce is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * yamce is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/
#include <QDebug>
#include <QMutex>
#include <QWaitCondition>
#include <QMessageBox>
#include <QDir>

#include "mainwindow.h"
#include "commandrailmap.h"
#include "commandman.h"
#include "fileman.h"
#include "nbtobject.h"
#include "regionfile.h"
#include "mapfile.h"
#include "world.h"
#include "railman.h"

static const bool DEBUG = false;

QMutex			railMutex;
QWaitCondition	railWait;

#undef errorMessage
#define errorMessage(__s) { \
	if (!errorMsgSent) \
		qCritical() << __s; \
	errorMsgSent = true; }

static const XYZ	neighbors[] ={
	XYZ(-1, -1, -1), XYZ(-1, -1, 0), XYZ(-1, -1, 1),
	XYZ(-1,  0, -1), XYZ(-1,  0, 0), XYZ(-1,  0, 1),
	XYZ(-1,  1, -1), XYZ(-1,  1, 0), XYZ(-1,  1, 1),
	XYZ( 0, -1, -1), XYZ( 0, -1, 0), XYZ( 0, -1, 1),
	XYZ( 0,  0, -1),                 XYZ( 0,  0, 1),
	XYZ( 0,  1, -1), XYZ( 0,  1, 0), XYZ( 0,  1, 1),
	XYZ( 1, -1, -1), XYZ( 1, -1, 0), XYZ( 1, -1, 1),
	XYZ( 1,  0, -1), XYZ( 1,  0, 0), XYZ( 1,  0, 1),
	XYZ( 1,  1, -1), XYZ( 1,  1, 0), XYZ( 1,  1, 1)
};


inline void RailMan::updateStats(const XYZ &xyz, const NBTObject* obj)
{
	if (xyz.x < m_railStats.farthestWest)
		m_railStats.farthestWest = xyz.x;
	if (xyz.x > m_railStats.farthestEast)
		m_railStats.farthestEast = xyz.x;
	if (xyz.z < m_railStats.farthestNorth)
		m_railStats.farthestNorth = xyz.z;
	if (xyz.z > m_railStats.farthestSouth)
		m_railStats.farthestSouth = xyz.z;
	m_railStats.railsSeen++;
	if (*obj->d_string == "minecraft:rail")
		m_railStats.normalRailsCount++;
	else if (*obj->d_string == "minecraft:powered_rail")
		m_railStats.poweredRailsCount++;
	else if (*obj->d_string == "minecraft:detector_rail")
		m_railStats.detectorRailsCount++;
	else if (*obj->d_string == "minecraft:activator_rail")
		m_railStats.activatorRailsCount++;
}

RailMan::RailMan(MainWindow* parent)
	: QObject(parent),
	  m_mainWindow(parent),
	  m_fileMan(nullptr),
	  m_world(nullptr),
	  m_running(false),
	  errorMsgSent(false),
	  m_railManThread(nullptr),
	  m_railMapThread(nullptr)
{
}

RailMan::~RailMan()
{
	if (m_railManThread)
		m_railManThread->deleteLater();
	if (m_railMapThread)
		m_railMapThread->deleteLater();
}

void RailMan::buildRailPoints()
{
	qInfo() << "RailMan:buildRailPoints";
	if (!m_world)
		return;
	QList<XYZ*> start = m_world->m_yamceDoc.m_mapSettings.m_railStartPoints;
	m_running = true;
	m_railStats.clear();
	m_foundRailPoints.clear();
	m_railSearchTree.clear();
	foreach (XYZ* xyz, start)
		m_railSearchTree.append(*xyz);

	m_railStats.farthestEast = m_railStats.farthestWest = m_railSearchTree.at(0).x;
	m_railStats.farthestNorth = m_railStats.farthestSouth = m_railSearchTree.at(0).z;
	if (m_railManThread)
		delete m_railManThread;
	m_railManThread = new RailManThread();
	m_railManThread->setup(this);
	connect(m_railManThread, SIGNAL(railManThreadDone()), this, SLOT(onRailManThreadDone()));
	m_railManThread->start();
}

void RailMan::onRailManThreadDone()
{
	m_running = false;
	saveRailsData();
	emit(railManDone());
	delete m_railManThread;
	m_railManThread = nullptr;
}

void RailManThread::setup(RailMan* railMan)
{
	m_railMan = railMan;
	m_fileMan = railMan->m_fileMan;
	m_world = railMan->m_world;
	m_railSearchTree = &railMan->m_railSearchTree;
	m_foundRailPoints = &railMan->m_foundRailPoints;
}


void RailManThread::run()
{
	forever {
		int i;
		for (i=m_railSearchTree->size()-1; i>=0; i--) {
			XYZ xyz=m_railSearchTree->at(i);
//			if (xyz.x == -513)
//				qDebug() << "-513";
			RegionFile* rf = m_world->findRegion(blockToRegion(xyz));
			if (!rf) {
				m_railSearchTree->removeAt(i);
				break;
			}
			if (rf->isLoaded()) {
				rf->setLastUse();
				NBTObject* obj;
				if ((obj = m_railMan->getRail(rf, xyz)) != nullptr) {
					if (!m_foundRailPoints->contains(xyz)) {
						m_foundRailPoints->append(xyz);
						m_railMan->updateStats(xyz, obj);
						if (DEBUG)
							qDebug() << "RailMan: found!" << xyz;
					}
					m_railMan->m_railSearchTree.removeAt(i);
					searchNeighbors(xyz);
					break;
				} else {
					m_railMan->m_railSearchTree.removeAt(i);
				}
			} else {
				if (DEBUG)
					qDebug() << "RailMan: need file" << rf->getFileName() << "for" << xyz;
				m_fileMan->readFile(rf);
			}
		}
		if (m_railMan->m_railSearchTree.size() == 0) {
			qInfo() << "RailMan: Done! foundCount="<< m_railMan->m_foundRailPoints.size();
			emit (railManThreadDone());
			return;			// done!
		}
		if (i < 0) {
			if (DEBUG)
				qDebug() << "RailMan: Wait.  foundCount=" << m_railMan->m_foundRailPoints.size() << "searchSize=" << m_railMan->m_railSearchTree.size();
			railMutex.lock();
			railWait.wait(&railMutex);
			railMutex.unlock();
			continue;
		}
		if (DEBUG)
			qDebug() << "RailMan: Bottom of forever. i=" << i << "foundCount=" << m_railMan->m_foundRailPoints.size() << "searchSize=" << m_railMan->m_railSearchTree.size();
	}
}

void RailManThread::searchNeighbors(const XYZ& xyz)
{
	XYZ newxyz;
	for (int i=0; i<26; i++) {
		newxyz = xyz;
		newxyz += neighbors[i];
		if (m_railMan->m_foundRailPoints.contains(newxyz))
			continue;
		if (m_railMan->m_railSearchTree.contains(newxyz))
			continue;
		m_railMan->m_railSearchTree.append(newxyz);
	}
}

void RailMan::fileLoaded(NBTFile* file)
{
	if (!m_running)
		return;
	RegionFile* rf = dynamic_cast<RegionFile*>(file);
	if (!rf)
		return;
	if (DEBUG)
		qDebug() << "RailMan: file loaded" << file->getFileName();
	railWait.wakeAll();
}

NBTObject* RailMan::getRail(RegionFile* rf, const XYZ& xyz)
{
	NBTObject* obj = rf->getBlockPalette(xyz);
	if (!obj)
		return(nullptr);
	NBTObject* nobj = obj->findChild("*:Name");
	if (!nobj)
		return(nullptr);
	if (*nobj->d_string == "minecraft:rail")
		return(nobj);
	if (*nobj->d_string == "minecraft:powered_rail")
		return(nobj);
	if (*nobj->d_string == "minecraft:detector_rail")
		return(nobj);
	if (*nobj->d_string == "minecraft:activator_rail")
		return(nobj);
	return(nullptr);
}

void RailMan::saveRailsData()
{
	QDir dir = m_world->m_yamceDoc.getYamceDir(m_world, true);
	QString filename = dir.absoluteFilePath("yamceRails.dat");
	QFile file(filename);
	bool res = file.open(QIODevice::WriteOnly);
	if (!res) {
		QString msg = QString("Failed to open %1 for writing").arg(filename);
		QMessageBox::critical(m_mainWindow, "Error", msg, QMessageBox::NoButton);
		return;
	}
	QDataStream out(&file);
	out << m_foundRailPoints;
	out << m_railStats.farthestWest;
	out << m_railStats.farthestEast;
	out << m_railStats.farthestNorth;
	out << m_railStats.farthestSouth;
	out << m_railStats.railsSeen;
	out << m_railStats.normalRailsCount;
	out << m_railStats.poweredRailsCount;
	out << m_railStats.detectorRailsCount;
	out << m_railStats.activatorRailsCount;
	file.close();
}

void RailMan::loadRailsData()
{
	QDir dir = m_world->m_yamceDoc.getYamceDir(m_world, false);
	QString filename = dir.absoluteFilePath("yamceRails.dat");
	QFile file(filename);
	m_railStats.clear();
	m_foundRailPoints.clear();
	bool res = file.open(QIODevice::ReadOnly);
	if (!res) {
		qInfo() << "RailMan::loadRailsData no rail data";
//		QString msg = QString("Failed to open %1 for reading").arg(filename);
//		QMessageBox::critical(m_mainWindow, "Error", msg, QMessageBox::NoButton);
		return;
	}
	QDataStream in(&file);
	in >> m_foundRailPoints;
	in >> m_railStats.farthestWest;
	in >> m_railStats.farthestEast;
	in >> m_railStats.farthestNorth;
	in >> m_railStats.farthestSouth;
	in >> m_railStats.railsSeen;
	in >> m_railStats.normalRailsCount;
	in >> m_railStats.poweredRailsCount;
	in >> m_railStats.detectorRailsCount;
	in >> m_railStats.activatorRailsCount;
	file.close();
}

RailStats& RailStats::operator=(const RailStats& other)
{
	this->farthestWest = other.farthestWest;
	this->farthestEast = other.farthestEast;
	this->farthestNorth = other.farthestNorth;
	this->farthestSouth = other.farthestSouth;
	this->railsSeen = other.railsSeen;
	this->normalRailsCount = other.normalRailsCount;
	this->poweredRailsCount = other.poweredRailsCount;
	this->detectorRailsCount = other.detectorRailsCount;
	this->activatorRailsCount = other.activatorRailsCount;
	return(*this);
}

static const QString tdright = "<td align=\"right\">";
static const QString tde = "</td>";
QString	RailMan::getTextStats()
{
	QString ret = "<table>";
	ret += QString("<tr><td colspan=\"3\">Farthest:</td></tr>");
	ret += QString("<tr><td>&nbsp;</td><td>%1</td><td>&nbsp;</td></tr>").arg(m_railStats.farthestNorth);
	ret += QString("<tr><td>%1</td><td>&nbsp;</td><td>%2</td></tr>").arg(m_railStats.farthestWest).arg(m_railStats.farthestEast);
	ret += QString("<tr><td>&nbsp;</td><td>%1</td><td>&nbsp;</td></tr>").arg(m_railStats.farthestSouth);
	ret += QString("<tr><td>&nbsp;</td></tr></table><table>");
	ret += QString("<tr><td><u>Total rails:</u></td>%1<u>%3</u>%2</tr>").arg(tdright, tde).arg(m_railStats.railsSeen);
	ret += QString("<tr><td>Rails:</td>%1%3%2</tr>").arg(tdright, tde).arg(m_railStats.normalRailsCount);
	ret += QString("<tr><td>Powered Rails:</td>%1%3%2</tr>").arg(tdright, tde).arg(m_railStats.poweredRailsCount);
	ret += QString("<tr><td>Detector Rails:</td>%1%3%2</tr>").arg(tdright, tde).arg(m_railStats.detectorRailsCount);
	ret += QString("<tr><td>Activator Rails:</td>%1%3%2</tr>").arg(tdright, tde).arg(m_railStats.activatorRailsCount);
	ret += QString("</table>");
	return(ret);
}

void RailMan::onUpdateMinecraftMaps()
{
	if (m_railMapThread)
		delete m_railMapThread;
	m_railMapThread = new RailMapThread();
	connect(m_railMapThread, SIGNAL(railMapStatus(int,int)), this, SLOT(onRailMapStatus(int,int)));
	connect(m_railMapThread, SIGNAL(railMapMap(NBTFile*, QList<QPoint>*)), this, SLOT(onRailMapMap(NBTFile*, QList<QPoint>*)));
	connect(m_railMapThread, SIGNAL(railMapDone()), this, SLOT(onRailMapDone()));
	m_railMapThread->setup(this);
	m_railMapThread->start();
}

void RailMan::onRailMapStatus(int done, int total)
{
	m_mapWriteProgress.m_text = "Saved %1 of %2 maps";
	m_mapWriteProgress.m_total = total;
	m_mapWriteProgress.m_done = done;
	m_mainWindow->setProgress(&m_mapWriteProgress);
}

void RailMan::onRailMapMap(NBTFile* file, QList<QPoint>* points)
{
	CommandRailMap* cmr = new CommandRailMap();
	cmr->setPoints(points);
	cmr->setFile(file);
	cmr->setColor(m_world->m_yamceDoc.m_railColor);
	m_mainWindow->getCommandMan()->doCommand(cmr);
}

void RailMan::onRailMapDone()
{
	emit(railManDone());
	m_railMapThread->deleteLater();
	m_railMapThread = nullptr;
}

void RailMapThread::setup(RailMan* railMan)
{
	m_railMan = railMan;
	m_world = m_railMan->m_world;
}

void RailMapThread::run()
{
	QThread t;

	QPoint railPoint;
	QPoint center = QPoint(64, 64);
	QList<QPoint>*	points = new QList<QPoint>();
	Group* maps = m_world->getGroup(wgMaps);
	int done = 0;
	emit (railMapStatus(done, maps->getFiles().size()));
	foreach (NBTFile* file, maps->getFiles()) {
		MapFile* mf = (MapFile*)file;
		if (!mf->m_imageDisplayed)
			continue;

		int scaler = 1 << mf->m_scale;
		QRect	rect = QRect(mf->m_pos.x()-(64*scaler), mf->m_pos.y()-(64*scaler), 128*scaler, 128*scaler);
		foreach(XYZ xyz, m_railMan->m_foundRailPoints) {
			if (xyz.inside(rect)) {
				railPoint = QPoint(xyz.x-mf->m_pos.x(), xyz.z - mf->m_pos.y());
				railPoint /= scaler;
				railPoint += center;
				points->append(railPoint);
			}
		}
		if (!points->isEmpty()) {
			emit(railMapMap(file, points));		// ownership of points is transferred
			points = new QList<QPoint>();
		}
		done++;
		emit (railMapStatus(done, maps->getFiles().size()));
	}
	delete points;
	emit(railMapDone());
}
