/******************************************************************************
 * railman.h - Find and Manage the rail locations from the world
 *
 * yamce - Yet Another Minecraft Editor
 * Copyright(c) 2018 - Dick Balaska, and BuckoSoft.
 *
 * yamce is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * yamce is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/
#ifndef RAILMAN_H
#define RAILMAN_H

#include <QPoint>
#include <QObject>
#include <QThread>

#include "xyz.h"
#include "progressset.h"

class FileMan;
class MainWindow;
class World;
class NBTFile;
class NBTObject;
class RailManThread;
class RailMapThread;
class RegionFile;

/// Track some statistics about our railroad
class RailStats
{
public:
	RailStats() {clear(); }
	RailStats& operator=(const RailStats& other);
	void	clear() {farthestWest = farthestEast = farthestNorth = farthestSouth = railsSeen =
					normalRailsCount = poweredRailsCount = detectorRailsCount = activatorRailsCount = 0;}

	int		farthestWest;
	int		farthestEast;
	int		farthestNorth;
	int		farthestSouth;
	int		railsSeen;
	int		normalRailsCount;
	int		poweredRailsCount;
	int		detectorRailsCount;
	int		activatorRailsCount;
};


/// Build a map of our railroad.
/// Starting at a point, hopefully with a rail type, follow any rails in every path and
/// build a list of known rails.
///
class RailMan : public QObject
{
	Q_OBJECT
public:
	explicit RailMan(MainWindow* parent);
	virtual ~RailMan();

	void		setFileMan(FileMan* fm) { m_fileMan = fm; }
	void		setWorld(World* world) { m_world = world; loadRailsData(); }
	NBTObject*	getRail(RegionFile* rf, const XYZ& xyz);
	QString		getTextStats();


	const QList<XYZ>&	getRails() {return (m_foundRailPoints); }
	const RailStats&	getRailStats() { return(m_railStats); }
signals:
	void	railManDone();

public slots:
	void	buildRailPoints();
	void	fileLoaded(NBTFile* file);
	void	onRailManThreadDone();
	void	onUpdateMinecraftMaps();

private slots:
	void	onRailMapStatus(int done, int total);
	void	onRailMapMap(NBTFile* file, QList<QPoint>* points);
	void	onRailMapDone();

private:
	void	loadRailsData();
	void	saveRailsData();
	//void	checkNextPoint();
	void	updateStats(const XYZ& xyz, const NBTObject* obj);

//	void parseRegionFile(RegionFile* rf);
//	bool scanBlock(int id, int paletteCount, NBTObject* blocks, QList<QPoint>& points);
//	void addRail(RegionFile* rf, NBTObject* chunkRoot, QList<QPoint>& points);

	MainWindow*		m_mainWindow;
	FileMan*		m_fileMan;
	World*			m_world;
	QList<XYZ>		m_railSearchTree;
	QList<XYZ>		m_foundRailPoints;
	bool			m_running;
	RailStats		m_railStats;
	QString			m_textStats;
	bool			errorMsgSent;
	RailManThread*	m_railManThread;
	RailMapThread*	m_railMapThread;
	ProgressSet		m_mapWriteProgress;

	friend class RailManThread;
	friend class RailMapThread;
};

class RailManThread : public QThread
{
	Q_OBJECT
protected:
	virtual void run();

signals:
	void	railManThreadDone();

private:
	void		setup(RailMan* railMan);
	void		searchNeighbors(const XYZ& xyz);

	RailMan*		m_railMan;
	FileMan*		m_fileMan;
	World*			m_world;
	QList<XYZ>*		m_railSearchTree;
	QList<XYZ>*		m_foundRailPoints;

	friend class RailMan;
};

class RailMapThread : public QThread
{
	Q_OBJECT

protected:
	virtual void run();

signals:
	void	railMapStatus(int done, int total);
	void	railMapMap(NBTFile* file, QList<QPoint>* points);
	void	railMapDone();

private:
	void	setup(RailMan* railMan);
	RailMan*		m_railMan;
	World*			m_world;

	friend class RailMan;
};

#endif // RAILMAN_H
