/******************************************************************************
 * regionwidget.cpp - The central widget that displays the map of the regions
 *
 * yamce - Yet Another Minecraft Editor
 * Copyright(c) 2018 - Dick Balaska, and BuckoSoft.
 *
 * yamce is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * yamce is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/
#include <QDebug>
#include <QLabel>
#include <QPainter>
#include <QPaintEvent>
#include <QVBoxLayout>
#include <QComboBox>
#include <QScrollArea>

#include "regionfile.h"
#include "world.h"
#include "mainwindow.h"
#include "regionwidget.h"

static const bool DEBUG = false;

RegionWidget::RegionWidget(MainWindow* mainWindow, QWidget* parent)
	: QWidget(parent)
{
	QVBoxLayout* layout = new QVBoxLayout();
	QComboBox* cb = new QComboBox();
	QStringList sl;
	sl << "Regions" << "Chunks" << "Position";
	cb->addItems(sl);
	layout->addWidget(cb);
	m_regionMapScrollArea = new RegionMapScrollArea(mainWindow);
	layout->addWidget(m_regionMapScrollArea);
	this->setLayout(layout);
}

RegionMapWidget::RegionMapWidget(MainWindow* , QScrollArea* parent)
	: QWidget(parent),
	  m_world(nullptr),
	  m_regionSizeX(0),
	  m_regionSizeY(0),
	  m_zoom(1.0)
{
}

void RegionMapWidget::setWorld(World* newWorld)
{
	m_world = newWorld;
	m_regionSizeX = m_world->m_maxRegionX - m_world->m_minRegionX +1;
	m_regionSizeY = m_world->m_maxRegionZ - m_world->m_minRegionZ +1;
	qDebug() << "m_regionX/Y=" << m_regionSizeX << "/" << m_regionSizeY;
}

void RegionMapWidget::paintEvent(QPaintEvent*)
{
	//qDebug() << "paintEvent";
	if (!m_world) {
		qDebug() << "paint - no world";
		return;
	}
	paintOverworldRegions();
}
void RegionMapWidget::paintOverworldRegions()
{
	QPainter p(this);
	int regionWidth = size().width() / m_regionSizeX;
	int regionHeight = size().height() / m_regionSizeY;
	if (regionWidth > regionHeight)
		regionWidth = regionHeight;
	else {
		regionHeight = regionWidth;
	}
	QFont font = p.font();
	QFontMetrics fm(font);
	int fmh = fm.height();
	for (int w=0; w<m_regionSizeX; w++) {
		for (int h=0; h<m_regionSizeY; h++) {
			int rx = m_world->m_minRegionX+w;
			int ry = m_world->m_minRegionZ+h;
			int wx = w*regionWidth;
			int hy = h*regionHeight;
			RegionFile* file = m_world->findRegion(rx, ry);
			if (file) {
				p.setPen(Qt::blue);
				p.drawRect(wx, hy, regionWidth, regionHeight);
				QString s = QString("%1/%2").arg(file->m_x).arg(file->m_z);
				p.drawText(wx+2, hy+fmh, s);
			}
			else {
				p.setPen(Qt::red);
				p.drawLine(wx, hy, wx+regionWidth, hy+regionHeight);
				p.drawLine(wx, hy+regionHeight, wx+regionWidth, hy);
			}
		}
	}
}

void RegionMapWidget::resizeEvent(QResizeEvent *event)
{
	//qDebug() << "resizeEvent" << event->size();
	QWidget::resizeEvent(event);
}

void RegionMapWidget::mousePressEvent(QMouseEvent* event)
{
	int regionWidth = size().width() / m_regionSizeX;
	int regionHeight = size().height() / m_regionSizeY;
	if (regionWidth > regionHeight)
		regionWidth = regionHeight;
	else {
		regionHeight = regionWidth;
	}
	int x = event->x() / regionWidth  + m_world->m_minRegionX;
	int z = event->y() / regionHeight + m_world->m_minRegionZ;
	emit (selectRegion(x, z));
	if (DEBUG)
		qDebug() << "event x/y:" << event->x() << "/" << event->y() << "x/z" << x << "/" << z;
}

///////////////////////////////////////////////////////////////////////////////
RegionMapScrollArea::RegionMapScrollArea(MainWindow* mainWindow, QWidget* parent)
	: QScrollArea(parent)
{
	//setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOn);
	m_widget = new RegionMapWidget(mainWindow, this);
	this->setWidget(m_widget);
}

void RegionMapScrollArea::resizeEvent(QResizeEvent *event)
{
	//qDebug() << "RegionMapScrollArea:resizeEvent" << event->size();
	QWidget::resizeEvent(event);
	m_widget->resize(event->size()*m_widget->m_zoom);
}

void RegionMapScrollArea::zoomValueChanged(double value)
{
	m_widget->m_zoom = value;
	m_widget->resize(this->size()*m_widget->m_zoom);
}

void RegionMapScrollArea::explorerTypeChanged(int newExplorerType)
{
	m_widget->m_explorerType = newExplorerType;
	m_widget->resize(this->size()*m_widget->m_zoom);
}
