/******************************************************************************
 * resourcedock.cpp - The left widget in the MainWindow
 *
 * yamce - Yet Another Minecraft Editor
 * Copyright(c) 2018 - Dick Balaska, and BuckoSoft.
 *
 * yamce is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * yamce is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/
#include <QDebug>
#include <QVBoxLayout>

#include "mainwindow.h"
#include "world.h"
#include "nbtfile.h"
#include "mapfile.h"
#include "regionfile.h"
#include "treemodel.h"
#include "treeitem.h"
#include "treeview.h"
#include "resourcedropdown.h"
#include "statusbar.h"
#include "resourcedock.h"
#include "commandman.h"
#include "commandedit.h"
#include "commandrename.h"

static bool DEBUG	= false;

ResourceDock::ResourceDock(MainWindow *parent, Qt::WindowFlags flags)
	: QDockWidget(QString("Resources"), parent, flags),
	  m_mainWindow(parent),
	  m_treeView(nullptr),
	  //m_treeModel(nullptr),
	  m_world(nullptr),
	  m_currentFile(nullptr)
{
	this->setObjectName("Resources");
	QWidget* dwidget = new QWidget(this);
	QVBoxLayout* layout = new QVBoxLayout();

	m_dropdown = new ResourceDropDown(m_mainWindow);
	layout->addWidget(m_dropdown);
	m_treeView = new TreeView(this);
	m_treeView->sortByColumn(0);
	layout->addWidget(m_treeView);
	dwidget->setLayout(layout);
	this->setWidget(dwidget);
	connect(m_treeView, SIGNAL(doubleClicked(const QModelIndex&)), this, SLOT(nodeDoubleClicked(const QModelIndex&)));
}

ResourceDock::~ResourceDock()
{
//	m_treeView->updateColumnSizeHints();
//	if (m_treeModel)
//		delete m_treeModel;
}

void ResourceDock::closeEvent(QCloseEvent* event)
{
	m_treeView->updateColumnSizeHints();
	QDockWidget::closeEvent(event);
}

void ResourceDock::closing()
{
	m_treeView->updateColumnSizeHints();
}

void ResourceDock::captureWidgets()
{
	m_treeView->updateColumnSizeHints();
}

bool ResourceDock::setGroupDropdownValue(int value)
{
	bool doEmit = false;
	if (m_dropdown->currentIndex() == value)
		doEmit = true;					// we are setting it to the same value, it won't do the emit on its own.
	m_dropdown->setCurrentIndex(value);
	return(doEmit);
}

void ResourceDock::setupColumnWidths()
{
	m_treeView->setupColumnWidths();
}

void ResourceDock::setWorld(World* world)
{
	m_world = world;
	for (Group* g : m_world->getGroups()) {
		if (!g->getTreeModel())
			g->setTreeModel(new TreeModel(this));
	}
//	if (m_treeModel) {
//		delete m_treeModel;
//	}
//	m_treeModel = new TreeModel(this);
	TreeModel* tm = m_world->getGroup(wgMain)->getTreeModel();
	tm->addNBTModelData(world->getGroup(wgMain)->getFiles().first()->getObject(), "level.dat");
	m_treeView->setModel(tm);
	setupColumnWidths();
	//m_treeView->setSortingEnabled(true);
}

void ResourceDock::explorerTypeChanged(int newGroup)
{
	if (!m_world)
		return;
	if (DEBUG)
		qDebug() << "ResourceDock::explorerTypeChanged" << newGroup;
	m_currentGroup = newGroup;
	m_group = m_world->getGroup(newGroup);
	//QItemSelectionModel* oldModel = m_treeView->selectionModel();
	//m_treeModel = new TreeModel(this);
	TreeModel* tm = m_group->getTreeModel();
	m_treeView->setSortingEnabled(false);
	if (tm->isEmpty()) {
		switch (newGroup) {
		case wgMain: {
			NBTFile* file = m_group->getFiles().first();
			tm->addNBTModelData(file->getObject(), file->getFileName());
			break;
		}
		case wgVillages:
		case wgMaps:
		case wgPlayers: {
			if (m_group->getFiles().isEmpty())
				break;
			foreach(NBTFile* file, m_group->getFiles()) {
				tm->addNBTModelData(file->getObject(), file->getFileName());
			}
			break;
		}
		}
	}
	m_treeView->updateColumnSizeHints();
	m_treeView->setModel(tm);
	m_treeView->setupColumnWidths();
	//m_treeView->setSortingEnabled(true);
	//delete oldModel;

}

void ResourceDock::setRegionFile(NBTFile* file)
{
	if (DEBUG)
		qDebug() << "ResourceDock:setFile" << file->getFileName();
	QItemSelectionModel* oldModel = m_treeView->selectionModel();
	TreeModel* tm = new TreeModel(this);
	m_currentFile = file;
	RegionFile* rf = dynamic_cast<RegionFile*>(file);
	if (rf) {
		tm->addRegionModelData(rf);
		m_group->setTreeModel(tm);
//	} else {
//		m_treeModel->addNBTModelData(file->getObject(), file->getFileName());
	}
	m_treeView->updateColumnSizeHints();
	m_treeView->setModel(tm);
	m_treeView->setupColumnWidths();
	delete oldModel;

}

void ResourceDock::setMapFile(MapFile* mapFile)
{
	qDebug() << "ResourceDock:setMapFile" << mapFile->getFileName();
	TreeModel* tm = m_group->getTreeModel();
	for (int i=0; i<tm->rowCount(); i++) {
		QModelIndex qmi = tm->index(i, 0);
		QVariant data = tm->data(qmi);
		QString s = data.toString();
		if (s.startsWith("--"))
			s = s.mid(2);
		//qDebug() << "data:" << s;
		if (s == mapFile->getFileName()) {
			m_treeView->selectionModel()->select(qmi, QItemSelectionModel::ClearAndSelect);
			m_treeView->scrollTo(qmi, QAbstractItemView::PositionAtCenter);
			return;
		}

	}
}

void ResourceDock::addMap(MapFile* file)
{
	if (m_currentGroup == wgMaps)
		m_group->getTreeModel()->addNBTModelData(file->getObject(), file->getFileName());
}

void ResourceDock::fileLoaded(NBTFile* file)
{
	if (DEBUG)
		qDebug() << "ResourceDock:fileLoaded" << file->getFileName();
	TreeModel* tm = m_group->getTreeModel();
	if (DEBUG)
		qDebug() << "ResourceDock:qmi children=" << tm->rowCount();
	QModelIndex parent;
	RegionFile* rf = dynamic_cast<RegionFile*>(file);
	for (int i=0; i<tm->rowCount(); i++) {
		QModelIndex qmi = tm->index(i, 0);
		QVariant data = tm->data(qmi);
		QString s = data.toString();
		if (s.startsWith("--"))
			s = s.mid(2);
		//qDebug() << "data:" << s;
		if (s == file->getFileName()) {
			if (DEBUG)
				qDebug() << "ResourceDock replace file" << file->getFileName();
			TreeItem* ti;
			if (rf)
				ti = tm->addRegionModelData(rf, true);
			else
				ti = tm->addNBTModelData(file->getObject(), file->getFileName(), true);
			m_treeView->setSortingEnabled(false);
			tm->replaceRow(parent, i, ti);
			m_treeView->setSortingEnabled(true);
			return;
		}
	}
}

void ResourceDock::nodeDoubleClicked(const QModelIndex& index)
{
	NBTObject* obj = m_group->getTreeModel()->getTreeItem(index)->getObject();
	qDebug() << "ResourceDock::nodeDoubleClicked obj=" << obj;
}

NBTFile* ResourceDock::getFile(TreeItem* treeItem)
{
	TreeItem* parent = treeItem->parentItem();
	while (parent) {
		if (!parent->parentItem() || !parent->parentItem()->parentItem())	// stop going up before the hidden root
			break;
		parent = parent->parentItem();
	}
	if (!parent)
		return(nullptr);
	QString fname = parent->getName();

	for (Group* g : m_world->getGroups()) {
		for (NBTFile* nf : g->getFiles()) {
			if (nf->getFileName() == fname)
				return(nf);
		}
	}
	return(nullptr);
}

TreeItem* ResourceDock::getTreeItem(NBTFile* file)
{
	for (Group* g : m_world->getGroups()) {
		TreeItem* ti = g->getTreeModel()->getRootNode();
		for (TreeItem* child : ti->getChildren()) {
			if (file->getObject() == child->getObject())
				return(child);
		}
	}
	return(nullptr);
}
void ResourceDock::onDataEdited(const QModelIndex& index, const QVariant data)
{
	//qDebug() << "ResourceDock:onDataEdited";
//	m_mainWindow->getCommandMan()->editDone(index, data);
	if (index.column() == 2) {
		QVariant oldData = index.data();
		if (oldData == data)
			return;
		CommandEdit* ce = new CommandEdit();
		ce->setGroup(m_group);
		ce->setEditData(index, data);
		m_mainWindow->getCommandMan()->doCommand(ce);
	} else if (index.column() == 0) {
		QVariant oldData = index.data();
		if (oldData == data)
			return;
		CommandRename* cr = new CommandRename();
		cr->setGroup(m_group);
		cr->setEditData(index, data);
		m_mainWindow->getCommandMan()->doCommand(cr);
	}
}
