/******************************************************************************
 * syncman.cpp - Keep yamce synchronized with a running Minecraft world
 *
 * yamce - Yet Another Minecraft Editor
 * Copyright(c) 2019 - Dick Balaska, and BuckoSoft.
 *
 * yamce is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * yamce is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/
#include <QFileSystemWatcher>
#include <QDir>
#include <QFileInfo>
#include <QDebug>

#include "mainwindow.h"
#include "world.h"
#include "fileman.h"
#include "playerfile.h"
#include "mapfile.h"
#include "resourcedock.h"
#include "syncman.h"

const bool DEBUG = false;
SyncMan::SyncMan(MainWindow* mainWindow)
	: m_mainWindow(mainWindow),
	  m_world(nullptr),
	  m_fileSystemWatcher(nullptr)
{
	connect(&m_timer, SIGNAL(timeout()), this, SLOT(onTimeout()));
}

SyncMan::~SyncMan()
{
	m_timer.stop();
	if (m_fileSystemWatcher)
		delete m_fileSystemWatcher;
	while (!m_readQueue.isEmpty())
		delete m_readQueue.takeLast();
	m_mapsFileInfoList.clear();
}

void SyncMan::onSynchronize(bool checked)
{
	if (!checked) {
		if (!m_fileSystemWatcher)
			return;
		qInfo() << "SyncMan shutdown";
		m_timer.stop();
		if (m_fileSystemWatcher)
			delete m_fileSystemWatcher;
		while (!m_readQueue.isEmpty())
			delete m_readQueue.takeLast();
		m_fileSystemWatcher = nullptr;
		m_mapsFileInfoList.clear();
		return;
	}
	qInfo() << "SyncMan startup";
	if (m_fileSystemWatcher)
		delete m_fileSystemWatcher;
	m_mapsFileInfoList.clear();
	m_timer.start(100);
	m_world = m_mainWindow->getWorld();
	m_fileSystemWatcher = new QFileSystemWatcher();
	connect(m_fileSystemWatcher, SIGNAL(fileChanged(const QString&)), this, SLOT(onFileChanged(const QString&)));
	connect(m_fileSystemWatcher, SIGNAL(directoryChanged(const QString&)), this, SLOT(onDirectoryChanged(const QString&)));
	m_worldDir = m_mainWindow->getWorld()->getWorldDirectory();
	m_mapDir = m_worldDir + "/data";
	m_playerDir = m_worldDir + "/playerdata";
	m_fileSystemWatcher->addPath(m_worldDir);
	m_fileSystemWatcher->addPath(m_mapDir);
	m_fileSystemWatcher->addPath(m_playerDir);
	addMapsToWatch();
	m_lastSync = QDateTime::currentDateTime().addSecs(-1);
}

void SyncMan::onFileChanged(const QString& path)
{
	if (DEBUG)
		qDebug() << "onFileChanged:" << path;
	if (path.startsWith(m_mapDir)) {
		if (DEBUG)
			qDebug() << "onFileChanged check maps";
		for (QFileInfo fi : m_mapsFileInfoList) {
			if (fi.filePath() == path) {
				mapChanged(fi);
				return;
			}
		}
		qWarning() << "Unknown map changed" << path;
		return;
	}
	qInfo() << "Unknown file changed" << path;
}
QStringList dotDatFilter = {"*.dat"};

void SyncMan::onDirectoryChanged(const QString& path)
{
	if (path == m_playerDir) {
		updatePlayerPositions();
		m_lastSync = QDateTime::currentDateTime().addSecs(0);
		return;
	}
	if (DEBUG)
		qDebug() << "onDirectoryChanged:" << path;
	if (path == m_mapDir) {
		mapDirectoryChanged();
		m_lastSync = QDateTime::currentDateTime().addSecs(0);
		return;
	}
	QDir dir(path);
	QFileInfoList fil = dir.entryInfoList(dotDatFilter);
	for (QFileInfo fi : fil) {
		if (fi.lastModified() > m_lastSync) {
			if (fi.fileName() == "level.dat")
				continue;
			qDebug() << "File modified:" << fi.fileName();
		}
	}
	m_lastSync = QDateTime::currentDateTime().addSecs(-1);
}

QStringList mapFilters = {"map_*.dat"};

void SyncMan::addMapsToWatch()
{
	QDir dir(m_mapDir);
	QFileInfoList fil = dir.entryInfoList(mapFilters, QDir::Files, QDir::Time);
	QString s;
	for (QFileInfo fi : fil) {
		m_mapsFileInfoList.append(fi);
		s = fi.filePath();
		bool b = m_fileSystemWatcher->addPath(s);
		if (!b) {
			qWarning() << "Failed to monitor mapFile" << s;
			break;
		}
		if (m_world->m_yamceDoc.m_maxMapsWatched && m_mapsFileInfoList.size() > m_world->m_yamceDoc.m_maxMapsWatched) {
			qInfo() << "Max maps watched" << m_mapsFileInfoList.size() << "reached";
			break;
		}
	}
	qInfo() << "Maps watched" << m_mapsFileInfoList.size();
	//qDebug() << "addMapsToWatch" << m_mapsFileInfoList;
}


void SyncMan::updatePlayerPositions()
{
	if (DEBUG)
		qDebug() << "updatePlayerPositions";
	QDir dir(m_playerDir);
	QFileInfoList fil = dir.entryInfoList(dotDatFilter);
	for (QFileInfo fi : fil) {
		if (DEBUG)
			qDebug() << "updatePlayerPositions";
		if (fi.lastModified() > m_lastSync) {
			if (DEBUG)
				qDebug() << "SyncMan update player" << fi.fileName();
			PlayerFile*	pf = m_world->findPlayer(fi.fileName());
			if (!pf) {
				qWarning() << "SyncMan: Can't find player:" << fi.fileName();
				return;
			}
			queueDelayedRead(pf);
		}
	}
}

void SyncMan::mapDirectoryChanged()
{
	QDir dir(m_mapDir);
	QFileInfoList fil = dir.entryInfoList(mapFilters, QDir::Files, QDir::Time);
	QFileInfo fi;
	while (!fil.isEmpty()) {
		fi = fil.takeFirst();
		bool found = false;
		for (QFileInfo mfi : m_mapsFileInfoList) {
			if (fi == mfi) {
				found = true;
				break;
			}
		}
		if (!found)
			mapChanged(fi);
	}
}

void SyncMan::mapChanged(const QFileInfo& fileInfo)
{
	if (DEBUG)
		qDebug() << "mapChanged:" << fileInfo.fileName();
	MapFile* mf = m_world->findMap(fileInfo.fileName());
	if (!mf) {
		if (DEBUG)
			qDebug() << "mapChanged: new map!" << fileInfo.fileName();
		mf = new MapFile();
		mf->init(m_mapDir, fileInfo.fileName());
		m_world->getGroup(wgMaps)->append(mf);
		m_mainWindow->getFileMan()->readFile(mf);
		m_mainWindow->getResourceDock()->addMap(mf);
		if (m_mapsFileInfoList.size() > m_world->m_yamceDoc.m_maxMapsWatched) {
			QFileInfo fi = m_mapsFileInfoList.takeLast();
			bool b = m_fileSystemWatcher->removePath(fi.filePath());
			if (!b) {
				qWarning() << "mapChanged Failed to remove watch on" << fi.filePath();
			}
		}
		m_mapsFileInfoList.prepend(fileInfo);
		bool b = m_fileSystemWatcher->addPath(fileInfo.filePath());
		if (!b)
			qWarning() << "Failed to monitor mapFile" << fileInfo.path();
		return;
	}
	queueDelayedRead(mf);
}

void SyncMan::queueDelayedRead(NBTFile* file)
{
	if (m_mainWindow->getFileMan()->queuedFile(file)) {
		qDebug() << "queueDelayedRead: Already reading file" << file->getFileName();
		return;
	}
	for (DelayedRead* dr : m_readQueue) {
		if (dr->m_file == file)
			return;
	}
	DelayedRead* dr = new DelayedRead();
	dr->m_file = file;
	file->unload();
	m_readQueue.append(dr);
}

void SyncMan::onTimeout()
{
	for (DelayedRead* dr : m_readQueue) {
		dr->m_tick--;
		if (dr->m_tick <= 0) {
			m_readQueue.removeAll(dr);
			if (!m_mainWindow->getFileMan()->queuedFile(dr->m_file)) {
				m_mainWindow->getFileMan()->readFile(dr->m_file);
			}
			delete dr;
			onTimeout();
			return;
		}
	}
}
