/******************************************************************************
 * world.cpp - Describe one Minecraft world
 *
 * yamce - Yet Another Minecraft Editor
 * Copyright(c) 2018 - Dick Balaska, and BuckoSoft.
 *
 * yamce is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * yamce is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/
#include <QDebug>
#include <QString>
#include <QDir>
#include <QFileInfoList>
#include <QMessageBox>

#include "mainwindow.h"
#include "nbtfile.h"
#include "regionfile.h"
#include "mapfile.h"
#include "treemodel.h"
#include "playerfile.h"
#include "world.h"

static const bool DEBUG = false;

static QStringList uncompressedFiles = {"data/idcounts.dat"};

QStringList mainEntries = { "level.dat", "players", "maps", "villages", "Overworld Regions", "Nether Regions", "End Regions"};

Group::Group()
	: m_tree(nullptr)
{
}

Group::~Group()
{
	if (m_tree)
		delete m_tree;
}

World::World(MainWindow* mainWindow)
	: m_mainWindow(mainWindow),
	  m_serverLayout(false)
{
	for (int i=0; i<wgMAX; i++) {
		m_groups.append(new Group);
	}
}
World::~World()
{
	while (!m_groups.isEmpty()) {
		Group* wg = m_groups.takeFirst();

		while (!wg->m_files.isEmpty()) {
			NBTFile* file = wg->m_files.takeFirst();
			delete file;
		}
		delete wg;
	}
}

bool World::load(const QString& dirname, ErrorMsgReceiver* receiver)
{
	QString s_error = tr("Error");
	QDir dir(dirname);
	m_name = dir.dirName();
	m_directory = dirname;
	QFileInfoList filist = dir.entryInfoList();
	if (filist.isEmpty()) {
		QString s = tr("\"%1\" is empty").arg(dirname);
		QMessageBox::critical(m_mainWindow, s_error, s, QMessageBox::NoButton);
		return(false);
	}
	bool found = false;
	foreach (QFileInfo qfi, filist) {
		if (qfi.fileName() == "level.dat") {
			found = true;
			break;
		}
	}
	if (!found) {
		foreach (QFileInfo qfi, filist) {
			if (qfi.fileName() == "world" && qfi.isDir()) {
				found = true;
				m_serverLayout = true;
				break;
			}
		}

	}
	if (!found) {
		QString s = tr("\"%1\" is not a minecraft directory (no level.dat)").arg(dirname);
		QMessageBox::critical(m_mainWindow, s_error, s, QMessageBox::NoButton);
		return(false);
	}
	dir.cd(getWorldDirectory());
	NBTFile* nf = new NBTFile();
	nf->init(getWorldDirectory(), "level.dat");
	nf->load(receiver);
	//m_files.append(nf);
	m_groups.at(wgMain)->append(nf);
	bool res = dir.cd("playerdata");
	if (!res) {
		QString s = tr("Failed to cd to playerdata directory");
		QMessageBox::critical(m_mainWindow, s_error, s, QMessageBox::NoButton);
		return(false);
	}
	QStringList nameFilters;
	nameFilters << "*.dat";
	dir.setNameFilters(nameFilters);
	filist = dir.entryInfoList();
	if (DEBUG)
		qDebug() << "Processing" << filist.size() << "players";
	foreach (QFileInfo qfi, filist) {
		PlayerFile* pf = new PlayerFile();
		if (DEBUG)
			qDebug() << qfi.fileName();
		pf->init(dir.path(), qfi.fileName());
		m_groups.at(wgPlayers)->append(pf);
	}
	res = dir.cd("../data");
	if (!res) {
		QString s = tr("Failed to cd to data directory");
		QMessageBox::critical(m_mainWindow, s_error, s, QMessageBox::NoButton);
	}
	nameFilters.clear();
	nameFilters << "map_*.dat";
	dir.setNameFilters(nameFilters);
	filist = dir.entryInfoList();
	if (DEBUG)
		qDebug() << "Processing" << filist.size() << "maps";
	foreach (QFileInfo qfi, filist) {
		MapFile* mf = new MapFile();
		mf->init(dir.path(), qfi.fileName());
		m_groups.at(wgMaps)->append(mf);
	}
	nameFilters.clear();
	nameFilters << "villages*.dat";
	dir.setNameFilters(nameFilters);
	filist = dir.entryInfoList();
	if (DEBUG)
		qDebug() << "Processing" << filist.size() << "village(s)";
	foreach (QFileInfo qfi, filist) {
		NBTFile* nf = new NBTFile();
		nf->init(dir.path(), qfi.fileName());
		m_groups.at(wgVillages)->append(nf);
	}
	dir.cd(getWorldDirectory());
	res = dir.cd("region");
	if (!res) {
		QString s = tr("Failed to cd to region directory");
		QMessageBox::critical(m_mainWindow, s_error, s, QMessageBox::NoButton);
	}
	nameFilters.clear();
	nameFilters << "r.*.mca";
	dir.setNameFilters(nameFilters);
	filist = dir.entryInfoList();
	if (DEBUG)
		qDebug() << "Processing" << filist.size() << "regions";
	m_minRegionX = m_minRegionZ = m_maxRegionX = m_maxRegionZ = 0;
	foreach (QFileInfo qfi, filist) {
		RegionFile* rf = new RegionFile();
		rf->init(dir.path(), qfi.fileName());
		QStringList sl = qfi.fileName().split(".");
		rf->m_x = sl[1].toInt();
		rf->m_z = sl[2].toInt();
		//qDebug() << "file:" << qfi.fileName() << rf->m_x << rf->m_z;
		if (rf->m_x < m_minRegionX)
			m_minRegionX = rf->m_x;
		if (rf->m_x > m_maxRegionX)
			m_maxRegionX = rf->m_x;
		if (rf->m_z < m_minRegionZ)
			m_minRegionZ = rf->m_z;
		if (rf->m_z > m_maxRegionZ)
			m_maxRegionZ = rf->m_z;
		m_groups.at(wgOverworldRegions)->append(rf);
	}
	if (DEBUG)
		qDebug() << "Regions: min:" << m_minRegionX << "/" << m_minRegionZ << "max:" << m_maxRegionX << "/" << m_maxRegionZ;
	return(true);
}

bool World::save(ErrorMsgReceiver* receiver)
{
	foreach (Group* group, m_groups) {
		foreach (NBTFile* nf, group->m_files) {
			if (nf->isDirty())
				nf->save(receiver);
			nf->setDirty(false);
		}
	}
	return(true);
}

QString World::getMapWallFileName()
{
	QString s = m_yamceDoc.m_exportPngDir;
	if (s.isEmpty())
		s = QDir::homePath();
	s += "/";
	s += m_name;
	s = s.replace(' ', '_');
	s = s.append('_');
	s = s.append(QDateTime::currentDateTime().toString("yyyy-MM-dd_hh.mm.ss"));
	s = s.append(".png");
	return(s);
}

RegionFile* World::findRegion(int x, int z)
{
	if (m_groups.size() < wgOverworldRegions)
		return(nullptr);
	Group* group = m_groups.at(wgOverworldRegions);
	foreach(NBTFile* nf, group->m_files) {
		RegionFile* rf = (RegionFile*)nf;
		if (rf->m_x == x && rf->m_z == z)
			return(rf);
	}
	return(nullptr);
}

PlayerFile*	World::findPlayer(const QString &fileName)
{
	Group* group = getGroup(wgPlayers);
	for (NBTFile* file : group->m_files) {
		if (file->m_fileName == fileName)
			return(static_cast<PlayerFile*>(file));
	}
	return(nullptr);
}

MapFile*	World::findMap(const QString &fileName)
{
	Group* group = getGroup(wgMaps);
	for (NBTFile* file : group->m_files) {
		if (file->m_fileName == fileName)
			return(static_cast<MapFile*>(file));
	}
	return(nullptr);
}

Group* World::findGroup(NBTFile* file)
{
	for (Group* g : m_groups) {
		if (g->getFiles().contains(file))
			return(g);
	}
	qCritical() << "Can't find group for file" << file->getFileName();
	return(nullptr);
}
