/******************************************************************************
 * yamcedoc.cpp - Manage the yamce accoutrements to a World
 *
 * yamce - Yet Another Minecraft Editor
 * Copyright(c) 2018 - Dick Balaska, and BuckoSoft.
 *
 * yamce is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * yamce is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/
#include <QDebug>
#include <QSettings>
#include <QDir>

#include "world.h"
#include "mapfile.h"
#include "mainwindow.h"
#include "yamcedoc.h"


static const QString	s_yamce				("yamce");
static const QString	s_MapWall			("MapWall");
static const QString	s_LiveUpdater		("LiveUpdater");
static const QString	s_WindowSettings	("WindowSettings");

static const QString	s_excludedMaps		("excludedMaps");
static const QString	s_gridWidth			("gridWidth");
static const QString	s_drawGridLines		("drawGridLines");
static const QString	s_expandForGrid		("expandForGrid");
static const QString	s_drawRailLines		("drawRailLines");
static const QString	s_railMapColor		("railMapColor");
static const QString	s_railStartPoints	("railStartPoints");
static const QString	s_railPointSpaces	("railPointSpaces");
static const QString	s_resourceColumns	("resourceColumns");
static const QString	s_currentGroup		("currentGroup");
static const QString	s_currentZoom		("currentZoom");
static const QString	s_mapZoomDisplayed	("mapZoomDisplayed");
static const QString	s_mapImagesDisplayed("mapImagesDisplayed");
static const QString	s_maxMapsWatched	("maxMapsWatched");
static const QString	s_exportPngDir		("exportPngDir");

static const QString	s_mapSettingsDialogSettingsPos("mapSettingsDialogSettingsPos");
static const QString	s_mapSettingsDialogSettingsOn("mapSettingsDialogSettingsOn");
static const QString	s_railManDialogSettingsPos("railManDialogSettingsPos");
static const QString	s_railManDialogSettingsOn("railManDialogSettingsOn");
static const QString	s_consoleDialogSettingsPos("consoleDialogSettingsPos");
static const QString	s_consoleDialogSettingsOn("consoleDialogSettingsOn");
static const QString	s_prefsDialogSettingsPos("prefsDialogSettingsPos");
static const QString	s_prefsDialogSettingsOn("prefsDialogSettingsOn");

YamceDoc::YamceDoc()
{
	m_currentResourceGroup = wgMain;
	m_currentZoom = 1.0;
	m_maxMapsWatched = 0;
	m_railColor = 18;
}

#define wLoad(WW)	{ \
	qv = settings.value(s_##WW##Pos); \
	if (qv.isValid()) \
		m_##WW.m_pos = qv.toRect(); \
	qv = settings.value(s_##WW##On); \
	if (qv.isValid()) \
		m_##WW.m_shown = qv.toBool(); \
}

#define wSave(WW)	 { \
	settings.setValue(s_##WW##Pos, m_##WW.m_pos); \
	settings.setValue(s_##WW##On, m_##WW.m_shown); \
}

bool YamceDoc::load(MainWindow*, World* world)
{
	QDir dir = getYamceDir(world, false);
	QString filename = dir.absoluteFilePath("yamce.ini");
	QSettings settings(filename, QSettings::IniFormat);
	settings.beginGroup(s_MapWall);
	QVariant qv = settings.value(s_excludedMaps);
	if (qv.isValid()) {
		QStringList excludedMapList = qv.toStringList();
		Group* maps = world->getGroup(wgMaps);
		foreach (NBTFile* file, maps->getFiles()) {
			foreach (const QString& s, excludedMapList) {
				if (file->getFileName() == s) {
					MapFile* mf = (MapFile*)file;
					mf->setImageDisplayed(false);
				}
			}
		}
	}
	MapSettings& ms = world->m_yamceDoc.m_mapSettings;
	qv = settings.value(s_gridWidth);
	if (qv.isValid())
		ms.m_gridWidth = qv.toInt();
	qv = settings.value(s_drawGridLines);
	if (qv.isValid())
		ms.m_drawGridLines = qv.toBool();
	qv = settings.value(s_expandForGrid);
	if (qv.isValid())
		ms.m_expandForGrid = qv.toBool();
	qv = settings.value(s_drawRailLines);
	if (qv.isValid())
		ms.m_drawRailLines = qv.toBool();
	qv = settings.value(s_railPointSpaces);
	if (qv.isValid())
		ms.m_railPointSpaces = qv.toBool();
	qv = settings.value(s_railStartPoints);
	if (qv.isValid())
		ms.decodeRailStartPoints(qv.toStringList());
	qv = settings.value(s_railMapColor);
	if (qv.isValid())
		m_railColor = qv.toInt();
	settings.endGroup();

	settings.beginGroup(s_WindowSettings);
	wLoad(mapSettingsDialogSettings);
	wLoad(railManDialogSettings);
	wLoad(consoleDialogSettings);
	wLoad(prefsDialogSettings);

	qv = settings.value(s_resourceColumns);
	if (qv.isValid()) {
		QList<QVariant> vl = qv.toList();
		foreach(QVariant v, vl) {
			m_resourceColumnWidths.append(v);
		}
	}
	qv = settings.value(s_mapImagesDisplayed);
	if (qv.isValid()) {
		bool b = qv.toBool();
		ms.m_displayMapImages = b;
	}
	qv = settings.value(s_currentGroup);
	if (qv.isValid()) {
		int i = qv.toInt();
		m_currentResourceGroup = i;
	}
	qv = settings.value(s_currentZoom);
	if (qv.isValid()) {
		double z = qv.toDouble();
		if (z <= 0.0) {
			qWarning() << "Preference bogus zoom value" << z << "using 1.0 instead";
			z = 1.0;
		}
		m_currentZoom = z;
	}
	qv = settings.value(s_mapZoomDisplayed);
	if (qv.isValid()) {
		int i = qv.toInt();
		ms.m_mapZoomDisplayed = i;
	}
	qv = settings.value(s_exportPngDir);
	if (qv.isValid())
		m_exportPngDir = qv.toString();

	settings.endGroup();

	settings.beginGroup(s_LiveUpdater);
	qv = settings.value(s_maxMapsWatched);
	if (qv.isValid())
		m_maxMapsWatched = qv.toInt();
	settings.endGroup();
	return(true);
}

bool YamceDoc::save(MainWindow*, World* world)
{
	if (!world)
		return(false);
	QDir dir = getYamceDir(world, true);
	QString filename = dir.absoluteFilePath("yamce.ini");
	QSettings settings(filename, QSettings::IniFormat);
	settings.beginGroup(s_MapWall);
	Group* maps = world->getGroup(wgMaps);
	QStringList excludedMapList;
	foreach (NBTFile* file, maps->getFiles()) {
		MapFile* mf = (MapFile*)file;
		if (!mf->isImageDisplayed()) {
			excludedMapList.append(mf->getFileName());
		}
	}
	MapSettings& ms = world->m_yamceDoc.m_mapSettings;
	settings.setValue(s_excludedMaps, excludedMapList);
	settings.setValue(s_gridWidth, ms.m_gridWidth);
	settings.setValue(s_drawGridLines, ms.m_drawGridLines);
	settings.setValue(s_expandForGrid, ms.m_expandForGrid);
	settings.setValue(s_drawRailLines, ms.m_drawRailLines);
	settings.setValue(s_railPointSpaces, ms.m_railPointSpaces);
	settings.setValue(s_railStartPoints, ms.encodeRailStartPoints());
	settings.setValue(s_railMapColor, m_railColor);
	settings.endGroup();


	settings.beginGroup(s_WindowSettings);
	wSave(mapSettingsDialogSettings);
	wSave(railManDialogSettings);
	wSave(consoleDialogSettings);
	wSave(prefsDialogSettings);

	settings.setValue(s_resourceColumns, m_resourceColumnWidths);
	settings.setValue(s_currentGroup, m_currentResourceGroup);
	settings.setValue(s_currentZoom, m_currentZoom);
	settings.setValue(s_mapZoomDisplayed, ms.m_mapZoomDisplayed);
	settings.setValue(s_mapImagesDisplayed, ms.m_displayMapImages);
	settings.setValue(s_exportPngDir, m_exportPngDir);
	settings.endGroup();

	settings.beginGroup(s_LiveUpdater);
	settings.setValue(s_maxMapsWatched, m_maxMapsWatched);
	settings.endGroup();
	return(true);
}

QDir YamceDoc::getYamceDir(World* world, bool createDir)
{
	QDir dir = QDir(world->getBaseDirectory());
	if (createDir && !dir.exists(s_yamce)) {
		dir.mkdir(s_yamce);
	}
	dir.cd(s_yamce);
	return(dir);
}

MapSettings::~MapSettings()
{
	while (!m_railStartPoints.isEmpty())
		delete m_railStartPoints.takeLast();
}
QStringList MapSettings::encodeRailStartPoints()
{
	QStringList sl;
	foreach (XYZ* rp, m_railStartPoints) {
		sl.append(rp->toString());
	}
	return(sl);
}

void MapSettings::decodeRailStartPoints(const QStringList& in)
{
	m_railStartPoints.clear();
	foreach (QString s, in) {
		XYZ* rp = new XYZ();
		rp->fromString(s);
		m_railStartPoints.append(rp);
	}
}

